#!/usr/bin/env python

import rsgislib
import rsgislib.imageutils
import rsgislib.imagecalc
import rsgislib.rastergis

from osgeo import gdal

import numpy

HAVE_SKI_SEG = True
try:
    import skimage.segmentation
except ImportError:
    HAVE_SKI_SEG = False

import os


def perform_felsenszwalb_segmentation(
    input_img,
    output_img,
    gdalformat="KEA",
    no_data_val=0,
    tmp_dir="tmp",
    calc_stats=True,
    use_pca=False,
    n_pca_bands=3,
    pca_pxl_sample=100,
    scale=1,
    sigma=0.8,
    min_size=20,
):
    """
    A function to perform the Felsenszwalb segmentation algorithm from the
    scikit-image library
    (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

    :param input_img: input image file.
    :param output_img: output image file.
    :param gdalformat: output image file format.
    :param tmp_dir: temp DIR used to output PCA files
    :param calc_stats: calculate image pixel statistics, histogram and image
                       pyramids - note if you are not using a KEA file then the
                       format needs to support RATs for this option as histogram
                       and colour table are written to RAT.
    :param use_pca: if there are not 1 or 3 image bands in the input file then
                    you can use PCA to reduce the number of image bands.
    :param n_pca_bands: the number of principle components outputs from the
                        PCA - needs to be either 1 or 3.
    :param scale: scikit-image Felsenszwalb parameter: 'Free parameter.
                  Higher means larger clusters.'
    :param sigma: scikit-image Felsenszwalb parameter: 'Width of Gaussian
                  kernel used in preprocessing.'
    :param min_size: scikit-image Felsenszwalb parameter: 'Minimum component
                     size. Enforced using postprocessing.'

    """
    if not HAVE_SKI_SEG:
        raise rsgislib.RSGISPyException(
            "Have not been able to import skimage.segmentation check it is installed."
        )
    # Create output image
    rsgislib.imageutils.create_copy_img(
        input_img, output_img, 1, 0, gdalformat, rsgislib.TYPE_32UINT
    )

    gdalDS = gdal.Open(input_img, gdal.GA_ReadOnly)
    pcaImg = ""
    if use_pca and (n_pca_bands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        inImgBaseName = os.path.splitext(os.path.basename(input_img))[0]
        pcaImg = os.path.join(tmp_dir, inImgBaseName + "_pca.kea")
        pcaEigenValsFile = os.path.join(tmp_dir, inImgBaseName + "_eigen.mtxt")
        inDType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
        rsgislib.imagecalc.perform_image_pca(
            input_img,
            pcaImg,
            pcaEigenValsFile,
            1,
            pca_pxl_sample,
            "KEA",
            inDType,
            no_data_val,
            False,
        )
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif use_pca and (n_pca_bands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        inImgBaseName = os.path.splitext(os.path.basename(input_img))[0]
        pcaImg = os.path.join(tmp_dir, inImgBaseName + "_pca.kea")
        pcaEigenValsFile = os.path.join(tmp_dir, inImgBaseName + "_eigen.mtxt")
        inDType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
        rsgislib.imagecalc.perform_image_pca(
            input_img,
            pcaImg,
            pcaEigenValsFile,
            3,
            pca_pxl_sample,
            "KEA",
            inDType,
            no_data_val,
            False,
        )
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif use_pca:
        gdalDS = None
        raise rsgislib.RSGISPyException(
            "In the number of bands given the number of principle components."
        )

    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise rsgislib.RSGISPyException(
            "Input image should have either a single band or three (RGB). Consider performing PCA to reduce."
        )

    pxlValsArr = []
    multichannel = True
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand + 1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
        multichannel = True
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
        multichannel = False

    segResult = skimage.segmentation.felzenszwalb(
        pxlVals, scale=scale, sigma=sigma, min_size=min_size, multichannel=multichannel
    )
    segResult = segResult + 1

    gdalOutDS = gdal.Open(output_img, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None

    if calc_stats:
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )


def perform_quickshift_segmentation(
    input_img,
    output_img,
    gdalformat="KEA",
    no_data_val=0,
    tmp_dir="tmp",
    calc_stats=True,
    use_pca=False,
    pca_pxl_sample=100,
    ratio=1.0,
    kernel_size=5,
    max_dist=10,
    sigma=0,
    convert_to_lab=True,
    random_seed=42,
):
    """
    A function to perform the quickshift segmentation algorithm from the
    scikit-image library
    (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

    :param input_img: input image file.
    :param output_img: output image file.
    :param gdalformat: output image file format.
    :param tmp_dir: temp DIR used to output PCA files
    :param calc_stats: calculate image pixel statistics, histogram and image
                       pyramids - note if you are not using a KEA file then the
                       format needs to support RATs for this option as histogram
                       and colour table are written to RAT.
    :param use_pca: if there are not 3 image bands in the input file then you can
                    use PCA to reduce the number of image bands.
    :param ratio: scikit-image Quickshift parameter: 'Balances color-space proximity
                  and image-space proximity. Higher values give more weight to
                  color-space. (between 0 and 1)'
    :param kernel_size: scikit-image Quickshift parameter: 'Width of Gaussian kernel
                        used in smoothing the sample density. Higher means
                        fewer clusters.'
    :param max_dist: scikit-image Quickshift parameter: 'Cut-off point for
                     data distances. Higher means fewer clusters.'
    :param sigma: scikit-image Quickshift parameter: 'Width for Gaussian
                  smoothing as preprocessing. Zero means no smoothing.'
    :param convert_to_lab: scikit-image Quickshift parameter: 'Whether the input
                           should be converted to Lab colorspace prior to segmentation.
                           For this purpose, the input is assumed to be RGB.'
    :param random_seed: scikit-image Quickshift parameter: 'Random seed used
                        for breaking ties.'

    """
    if not HAVE_SKI_SEG:
        raise rsgislib.RSGISPyException(
            "Have not been able to import skimage.segmentation check it is installed."
        )
    # Create output image
    rsgislib.imageutils.create_copy_img(
        input_img, output_img, 1, 0, gdalformat, rsgislib.TYPE_32UINT
    )

    gdalDS = gdal.Open(input_img, gdal.GA_ReadOnly)
    pcaImg = ""
    if use_pca and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        inImgBaseName = os.path.splitext(os.path.basename(input_img))[0]
        pcaImg = os.path.join(tmp_dir, inImgBaseName + "_pca.kea")
        pcaEigenValsFile = os.path.join(tmp_dir, inImgBaseName + "_eigen.mtxt")
        inDType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
        rsgislib.imagecalc.perform_image_pca(
            input_img,
            pcaImg,
            pcaEigenValsFile,
            3,
            pca_pxl_sample,
            "KEA",
            inDType,
            no_data_val,
            False,
        )
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif use_pca:
        gdalDS = None
        raise rsgislib.RSGISPyException(
            "In the number of bands given the number of principle components."
        )

    if not (gdalDS.RasterCount == 3):
        gdalDS = None
        raise rsgislib.RSGISPyException(
            "Input image should have three bands (RGB). Consider performing PCA to reduce."
        )

    pxlValsArr = []
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand + 1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()

    segResult = skimage.segmentation.quickshift(
        pxlVals,
        ratio=ratio,
        kernel_size=kernel_size,
        max_dist=max_dist,
        return_tree=False,
        sigma=sigma,
        convert2lab=convert_to_lab,
        random_seed=random_seed,
    )
    segResult = segResult + 1

    gdalOutDS = gdal.Open(output_img, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None

    if calc_stats:
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )


def perform_slic_segmentation(
    input_img,
    output_img,
    gdalformat="KEA",
    no_data_val=0,
    tmp_dir="tmp",
    calc_stats=True,
    use_pca=False,
    n_pca_bands=3,
    pca_pxl_sample=100,
    n_segments=100,
    compactness=10.0,
    max_iter=10,
    sigma=0,
    spacing=None,
    convert_to_lab=None,
    enforce_connectivity=True,
    min_size_factor=0.5,
    max_size_factor=3,
    slic_zero=False,
):
    """
    A function to perform the slic segmentation algorithm from the
    scikit-image library
    (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

    :param input_img: input image file.
    :param output_img: output image file.
    :param gdalformat: output image file format.
    :param tmp_dir: temp DIR used to output PCA files
    :param calc_stats: calculate image pixel statistics, histogram and image
                       pyramids - note if you are not using a KEA file then the
                       format needs to support RATs for this option as histogram
                       and colour table are written to RAT.
    :param use_pca: if there are not 1 or 3 image bands in the input file then
                    you can use PCA to reduce the number of image bands.
    :param n_pca_bands: the number of principle components outputs from the
                        PCA - needs to be either 1 or 3.
    :param n_segments: scikit-image Slic parameter: 'The (approximate) number
                       of labels in the segmented output image.'
    :param compactness: scikit-image Slic parameter: 'Balances color proximity
                        and space proximity. Higher values give more weight to space
                        proximity, making superpixel shapes more square/cubic. In
                        SLICO mode, this is the initial compactness. This parameter
                        depends strongly on image contrast and on the shapes of
                        objects in the image. We recommend exploring possible values
                        on a log scale, e.g., 0.01, 0.1, 1, 10, 100, before refining
                        around a chosen value.'
    :param max_iter: scikit-image Slic parameter: 'Maximum number of iterations
                     of k-means.'
    :param sigma: scikit-image Slic parameter: 'Width of Gaussian smoothing kernel
                  for pre-processing for each dimension of the image. The same sigma
                  is applied to each dimension in case of a scalar value. Zero means
                  no smoothing. Note, that sigma is automatically scaled if it is
                  scalar and a manual voxel spacing is provided (see Notes section).'
    :param spacing: scikit-image Slic parameter: 'The voxel spacing along each image
                    dimension. By default, slic assumes uniform spacing (same voxel
                    resolution along z, y and x). This parameter controls the weights
                    of the distances along z, y, and x during k-means clustering.'
    :param convert_to_lab: scikit-image Slic parameter: 'Whether the input should be
                           converted to Lab colorspace prior to segmentation. The
                           input image must be RGB. Highly recommended.'
    :param enforce_connectivity: scikit-image Slic parameter: 'Whether the generated
                                 segments are connected or not'
    :param min_size_factor: scikit-image Slic parameter: 'Proportion of the minimum
                            segment size to be removed with respect to the supposed
                            segment size "depth:paramwidth*height/n_segments"'
    :param max_size_factor: scikit-image Slic parameter: 'Proportion of the maximum
                            connected segment size. A value of 3 works in most of
                            the cases.'
    :param slic_zero: scikit-image Slic parameter: 'Run SLIC-zero, the zero-parameter
                      mode of SLIC.'

    """
    if not HAVE_SKI_SEG:
        raise rsgislib.RSGISPyException(
            "Have not been able to import skimage.segmentation check it is installed."
        )
    # Create output image
    rsgislib.imageutils.create_copy_img(
        input_img, output_img, 1, 0, gdalformat, rsgislib.TYPE_32UINT
    )

    gdalDS = gdal.Open(input_img, gdal.GA_ReadOnly)
    pcaImg = ""
    if use_pca and (n_pca_bands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        inImgBaseName = os.path.splitext(os.path.basename(input_img))[0]
        pcaImg = os.path.join(tmp_dir, inImgBaseName + "_pca.kea")
        pcaEigenValsFile = os.path.join(tmp_dir, inImgBaseName + "_eigen.mtxt")
        inDType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
        rsgislib.imagecalc.perform_image_pca(
            input_img,
            pcaImg,
            pcaEigenValsFile,
            1,
            pca_pxl_sample,
            "KEA",
            inDType,
            no_data_val,
            False,
        )
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif use_pca and (n_pca_bands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        inImgBaseName = os.path.splitext(os.path.basename(input_img))[0]
        pcaImg = os.path.join(tmp_dir, inImgBaseName + "_pca.kea")
        pcaEigenValsFile = os.path.join(tmp_dir, inImgBaseName + "_eigen.mtxt")
        inDType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
        rsgislib.imagecalc.perform_image_pca(
            input_img,
            pcaImg,
            pcaEigenValsFile,
            3,
            pca_pxl_sample,
            "KEA",
            inDType,
            no_data_val,
            False,
        )
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif use_pca:
        gdalDS = None
        raise rsgislib.RSGISPyException(
            "In the number of bands given the number of principle components."
        )

    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise rsgislib.RSGISPyException(
            "Input image should have either a single band or three (RGB). Consider performing PCA to reduce."
        )

    pxlValsArr = []
    multichannel = False
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand + 1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
        multichannel = True
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
        multichannel = False

    segResult = skimage.segmentation.slic(
        pxlVals,
        n_segments=n_segments,
        compactness=compactness,
        max_iter=max_iter,
        sigma=sigma,
        spacing=spacing,
        multichannel=multichannel,
        convert2lab=convert_to_lab,
        enforce_connectivity=enforce_connectivity,
        min_size_factor=min_size_factor,
        max_size_factor=max_size_factor,
        slic_zero=slic_zero,
    )
    segResult = segResult + 1

    gdalOutDS = gdal.Open(output_img, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None

    if calc_stats:
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )


def perform_watershed_segmentation(
    input_img,
    in_markers_img,
    output_img,
    gdalformat="KEA",
    no_data_val=0,
    tmp_dir="tmp",
    calc_stats=True,
    use_pca=False,
    n_pca_bands=3,
    pca_pxl_sample=100,
    compactness=0,
    watershed_line=False,
):
    """
    A function to perform the watershed segmentation algorithm from the
    scikit-image library
    (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

    :param input_img: input image file.
    :param in_markers_img: input markers image file.
    :param output_img: output image file.
    :param gdalformat: output image file format.
    :param tmp_dir: temp DIR used to output PCA files
    :param calc_stats: calculate image pixel statistics, histogram and image
                       pyramids - note if you are not using a KEA file then the
                       format needs to support RATs for this option as histogram
                       and colour table are written to RAT.
    :param use_pca: if there are not 1 or 3 image bands in the input file then you
                    can use PCA to reduce the number of image bands.
    :param n_pca_bands: the number of principle components outputs from the
                        PCA - needs to be either 1 or 3.
    :param compactness: scikit-image Watershed parameter: 'Use compact watershed
                        with given compactness parameter. Higher values result in
                        more regularly-shaped watershed basins; Peer Neubert & Peter
                        Protzel (2014). Compact Watershed and Preemptive SLIC: On
                        Improving Trade-offs of Superpixel Segmentation
                        Algorithms. ICPR 2014'
    :param watershed_line: scikit-image Watershed parameter: 'If watershed_line is
                           True, a one-pixel wide line separates the regions obtained
                           by the watershed algorithm. The line has the label 0.'

    """
    if not HAVE_SKI_SEG:
        raise rsgislib.RSGISPyException(
            "Have not been able to import skimage.segmentation check it is installed."
        )

    # Create output image
    rsgislib.imageutils.create_copy_img(
        input_img, output_img, 1, 0, gdalformat, rsgislib.TYPE_32UINT
    )

    gdalMarkersDS = gdal.Open(in_markers_img, gdal.GA_ReadOnly)
    if gdalMarkersDS.RasterCount != 1:
        gdalMarkersDS = None
        raise rsgislib.RSGISPyException(
            "Markers image should only have one image band."
        )

    gdalMarkerBand = gdalMarkersDS.GetRasterBand(1)
    markerArr = gdalMarkerBand.ReadAsArray()

    gdalDS = gdal.Open(input_img, gdal.GA_ReadOnly)
    pcaImg = ""
    if use_pca and (n_pca_bands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        inImgBaseName = os.path.splitext(os.path.basename(input_img))[0]
        pcaImg = os.path.join(tmp_dir, inImgBaseName + "_pca.kea")
        pcaEigenValsFile = os.path.join(tmp_dir, inImgBaseName + "_eigen.mtxt")
        inDType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
        rsgislib.imagecalc.perform_image_pca(
            input_img,
            pcaImg,
            pcaEigenValsFile,
            1,
            pca_pxl_sample,
            "KEA",
            inDType,
            no_data_val,
            False,
        )
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif use_pca and (n_pca_bands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        inImgBaseName = os.path.splitext(os.path.basename(input_img))[0]
        pcaImg = os.path.join(tmp_dir, inImgBaseName + "_pca.kea")
        pcaEigenValsFile = os.path.join(tmp_dir, inImgBaseName + "_eigen.mtxt")
        inDType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
        rsgislib.imagecalc.perform_image_pca(
            input_img,
            pcaImg,
            pcaEigenValsFile,
            3,
            pca_pxl_sample,
            "KEA",
            inDType,
            no_data_val,
            False,
        )
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif use_pca:
        gdalDS = None
        raise rsgislib.RSGISPyException(
            "In the number of bands given the number of principle components."
        )

    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise rsgislib.RSGISPyException(
            "Input image should have either a single band or three (RGB). Consider performing PCA to reduce."
        )

    pxlValsArr = []
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand + 1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()

    segResult = skimage.segmentation.watershed(
        pxlVals, markerArr, compactness=compactness, watershed_line=watershed_line
    )
    if not watershed_line:
        segResult = segResult + 1

    gdalOutDS = gdal.Open(output_img, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None

    if calc_stats:
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )


def perform_random_walker_segmentation(
    input_img,
    in_markers_img,
    output_img,
    gdalformat="KEA",
    no_data_val=0,
    tmp_dir="tmp",
    calc_stats=True,
    use_pca=False,
    n_pca_bands=3,
    pca_pxl_sample=100,
    beta=130,
    mode="bf",
    tol=0.001,
    spacing=None,
):
    """
    A function to perform the random walker segmentation algorithm from the
    scikit-image library
    (http://scikit-image.org/docs/stable/api/skimage.segmentation.html).

    :param input_img: input image file.
    :param in_markers_img: input markers image file - markers must be uniquely numbered.
    :param output_img: output image file.
    :param gdalformat: output image file format.
    :param tmp_dir: temp DIR used to output PCA files
    :param calc_stats: calculate image pixel statistics, histogram and image
                       pyramids - note if you are not using a KEA file then the format
                       needs to support RATs for this option as histogram and colour
                       table are written to RAT.
    :param use_pca: if there are not 1 or 3 image bands in the input file then you
                    can use PCA to reduce the number of image bands.
    :param n_pca_bands: the number of principle components outputs from the
                        PCA - needs to be either 1 or 3.
    :param beta: scikit-image random_walker parameter: 'Penalization coefficient for
                 the random walker motion (the greater beta, the more difficult
                 the diffusion).'
    :param mode: scikit-image random_walker parameter: 'Mode for solving the linear
                 system in the random walker algorithm. Available options
                 {'cg_mg', 'cg', 'bf'}.'
            * 'bf' (brute force): an LU factorization of the Laplacian is computed.
               This is fast for small images (<1024x1024), but very slow and
               memory-intensive for large images (e.g., 3-D volumes).
            * 'cg' (conjugate gradient): the linear system is solved iteratively
               using the Conjugate Gradient method from scipy.sparse.linalg. This is
               less memory-consuming than the brute force method for large images,
               but it is quite slow.
            * 'cg_mg' (conjugate gradient with multigrid preconditioner): a
               preconditioner is computed using a multigrid solver, then the
               solution is computed with the Conjugate Gradient method. This mode
               requires that the pyamg module (http://pyamg.org/) is installed. For
               images of size > 512x512, this is the recommended (fastest) mode.
    :param tol: scikit-image random_walker parameter: 'tolerance to achieve when
                solving the linear system, in cg’ and ‘cg_mg’ modes.'
    :param spacing: scikit-image random_walker parameter: 'Spacing between voxels
                    in each spatial dimension. If None, then the spacing between
                    pixels/voxels in each dimension is assumed 1.'

    """
    if not HAVE_SKI_SEG:
        raise rsgislib.RSGISPyException(
            "Have not been able to import skimage.segmentation check it is installed."
        )

    # Create output image
    rsgislib.imageutils.create_copy_img(
        input_img, output_img, 1, 0, gdalformat, rsgislib.TYPE_32UINT
    )

    gdalMarkersDS = gdal.Open(in_markers_img, gdal.GA_ReadOnly)
    if gdalMarkersDS.RasterCount != 1:
        gdalMarkersDS = None
        raise rsgislib.RSGISPyException(
            "Markers image should only have one image band."
        )

    gdalMarkerBand = gdalMarkersDS.GetRasterBand(1)
    markerArr = gdalMarkerBand.ReadAsArray()

    gdalDS = gdal.Open(input_img, gdal.GA_ReadOnly)
    pcaImg = ""
    if use_pca and (n_pca_bands == 1) and (gdalDS.RasterCount > 1):
        # Perform PCA
        gdalDS = None
        inImgBaseName = os.path.splitext(os.path.basename(input_img))[0]
        pcaImg = os.path.join(tmp_dir, inImgBaseName + "_pca.kea")
        pcaEigenValsFile = os.path.join(tmp_dir, inImgBaseName + "_eigen.mtxt")
        inDType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
        rsgislib.imagecalc.perform_image_pca(
            input_img,
            pcaImg,
            pcaEigenValsFile,
            1,
            pca_pxl_sample,
            "KEA",
            inDType,
            no_data_val,
            False,
        )
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif use_pca and (n_pca_bands == 3) and (gdalDS.RasterCount > 3):
        # Perform PCA
        gdalDS = None
        inImgBaseName = os.path.splitext(os.path.basename(input_img))[0]
        pcaImg = os.path.join(tmp_dir, inImgBaseName + "_pca.kea")
        pcaEigenValsFile = os.path.join(tmp_dir, inImgBaseName + "_eigen.mtxt")
        inDType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
        rsgislib.imagecalc.perform_image_pca(
            input_img,
            pcaImg,
            pcaEigenValsFile,
            3,
            pca_pxl_sample,
            "KEA",
            inDType,
            no_data_val,
            False,
        )
        gdalDS = gdal.Open(pcaImg, gdal.GA_ReadOnly)
    elif use_pca:
        gdalDS = None
        raise rsgislib.RSGISPyException(
            "In the number of bands given the number of principle components."
        )

    if not ((gdalDS.RasterCount == 1) or (gdalDS.RasterCount == 3)):
        gdalDS = None
        raise rsgislib.RSGISPyException(
            "Input image should have either a single band or three (RGB). Consider performing PCA to reduce."
        )

    multichannel = True
    pxlValsArr = []
    if gdalDS.RasterCount > 1:
        for nBand in numpy.arange(gdalDS.RasterCount):
            gdalBand = gdalDS.GetRasterBand(int(nBand + 1))
            pxlValsArr.append(gdalBand.ReadAsArray())
        pxlVals = numpy.stack(pxlValsArr, axis=-1)
        multichannel = True
    else:
        gdalBand = gdalDS.GetRasterBand(1)
        pxlVals = gdalBand.ReadAsArray()
        multichannel = False

    segResult = skimage.segmentation.random_walker(
        pxlVals,
        markerArr,
        beta=beta,
        mode=mode,
        tol=tol,
        multichannel=multichannel,
        spacing=spacing,
    )
    segResult = segResult + 1

    gdalOutDS = gdal.Open(output_img, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(segResult)
    gdalOutDS = None
    gdalDS = None

    if calc_stats:
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )



#!/usr/bin/env python
############################################################################
#  leastcostpath.py
#
#  Copyright 2017 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Calculate the least cost path across a raster.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 17/08/2017
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

from osgeo import gdal
from osgeo import osr
import numpy


def _coord2pixelOffset(rasterfn, x, y):
    raster = gdal.Open(rasterfn)
    geotransform = raster.GetGeoTransform()
    originX = geotransform[0]
    originY = geotransform[3]
    pixelWidth = geotransform[1]
    pixelHeight = geotransform[5]
    xOffset = int((x - originX) / pixelWidth)
    yOffset = int((y - originY) / pixelHeight)
    return xOffset, yOffset


def _createPath(CostSurfaceImg, costSurfaceArray, startCoord, stopCoord):
    import skimage.graph

    # coordinates to array index
    startCoordX = startCoord[0]
    startCoordY = startCoord[1]
    startIndexX, startIndexY = _coord2pixelOffset(
        CostSurfaceImg, startCoordX, startCoordY
    )

    stopCoordX = stopCoord[0]
    stopCoordY = stopCoord[1]
    stopIndexX, stopIndexY = _coord2pixelOffset(CostSurfaceImg, stopCoordX, stopCoordY)

    # create path
    indices, weight = skimage.graph.route_through_array(
        costSurfaceArray,
        (startIndexY, startIndexX),
        (stopIndexY, stopIndexX),
        geometric=True,
        fully_connected=True,
    )
    indices = numpy.array(indices).T
    path = numpy.zeros_like(costSurfaceArray)
    path[indices[0], indices[1]] = 1
    return path


def _array2raster(newRasterfn, rasterfn, array, outFormat):
    raster = gdal.Open(rasterfn)
    geotransform = raster.GetGeoTransform()
    originX = geotransform[0]
    originY = geotransform[3]
    pixelWidth = geotransform[1]
    pixelHeight = geotransform[5]
    cols = array.shape[1]
    rows = array.shape[0]

    driver = gdal.GetDriverByName(outFormat)
    outRaster = driver.Create(newRasterfn, cols, rows, 1, gdal.GDT_Byte)
    outRaster.SetGeoTransform((originX, pixelWidth, 0, originY, 0, pixelHeight))
    outband = outRaster.GetRasterBand(1)
    outband.WriteArray(array)
    outRasterSRS = osr.SpatialReference()
    outRasterSRS.ImportFromWkt(raster.GetProjectionRef())
    outRaster.SetProjection(outRasterSRS.ExportToWkt())
    outband.FlushCache()
    outRaster = None
    raster = None


def perform_least_cost_path_calc(
    cost_surface_img: str,
    output_img: str,
    start_coord: list,
    stop_coord: list,
    gdalformat: str = "KEA",
    cost_img_band: int = 1,
):
    """
    Calculates least cost path for a raster surface from start coord to stop coord:

    Version of code from:
    https://pcjericks.github.io/py-gdalogr-cookbook/raster_layers.html#create-least-cost-path

    :param cost_surface_img: Input image to calculate cost path from
    :param output_img: Output image
    :param start_coord: Start coordinate (e.g., (263155.9, 291809.1))
    :param stop_coord: End coordinate (e.g., (263000.1, 292263.7))
    :param gdalformat: GDAL format (default=KEA)
    :param cost_img_band: Band in input image to use for cost analysis (default=1)

    """
    # Read gdal raster as array.
    gdalRasterDS = gdal.Open(cost_surface_img)
    costSurfaceArray = gdalRasterDS.GetRasterBand(cost_img_band).ReadAsArray()

    # Creates path array
    pathArray = _createPath(cost_surface_img, costSurfaceArray, start_coord, stop_coord)

    # Converts path array to raster
    _array2raster(output_img, cost_surface_img, pathArray, gdalformat)



#! /usr/bin/env python

############################################################################
# Copyright (c) 2016 Dr. Peter Bunting, Aberystwyth University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#
# Purpose:  A command to use the rsgislib functions to provide a command line
#           tool for calculating image stats and pyramids
#
# Author: Pete Bunting
# Email: pfb@aber.ac.uk
# Date: 20/04/16
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
#############################################################################

import argparse
import os
import rsgislib.imageutils
import rsgislib.rastergis


if __name__ == '__main__':
    """
    The command line user interface
    """
    parser = argparse.ArgumentParser(prog='rsgiscalcimgstats.py', description='''Calculate image statistics and pyramids''')

    parser.add_argument("-i", "--image", type=str, required=True, help='''The image on which stats and pyramids are to calculated.''')
    parser.add_argument("-t", "--thematic", action='store_true', default=False, help='''The image is thermatic, calc stats and pyramids accordingly.''')    
    parser.add_argument("-p", "--nopyramids", action='store_true', default=False, help='''Do not calculate image pyramids.''')
    parser.add_argument("-n", "--nodata", type=float, required=True, default=0.0, help='''No data value for the input image for when calculating stats''')
    parser.add_argument("-c", "--noclrtab", action='store_true', default=False, help='''Do not add a colour table (only for thematic images)''')    

    # Call the parser to parse the arguments.
    args = parser.parse_args()
    
    input_img = args.image
    pyramids = not args.nopyramids        
    
    if args.thematic:
        clr_tab = not args.noclrtab
        ignore_zeros = False
        if args.nodata == 0.0:
            ignore_zeros = True
        if os.path.splitext(input_img)[1].lower() == "kea":
            rsgislib.rastergis.pop_rat_img_stats(clumps_img=input_img, add_clr_tab=clr_tab, calc_pyramids=pyramids, ignore_zero=ignore_zeros)
        else:
            rsgislib.imageutils.pop_thmt_img_stats(input_img, add_clr_tab=clr_tab, calc_pyramids=pyramids, ignore_zero=ignore_zeros)
    else:
        rsgislib.imageutils.pop_img_stats(input_img, use_no_data=True, no_data_val=args.nodata, calc_pyramids=pyramids)




#!/usr/bin/env python
############################################################################
#  imagecomp.py
#
#  Copyright 2017 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities for creating image composites
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 04/10/2017
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

import os
import shutil

import osgeo.gdal
import rios.rat
import numpy

import rsgislib
import rsgislib.imagecalc
import rsgislib.imagecalc.calcindices
import rsgislib.imageutils
import rsgislib.rastergis



def check_build_ls8_ls9_vrts(input_imgs, output_dir):
    """
    A function which checks for Landsat 8 (LS8) and Landsat 9 (LS9) images
    in the input list and creates a band subset (i.e., removes the coastal
    (1) band). This function should be used ahead of create_max_ndvi_ndwi_composite
    when combining LS8 or LS8 with earlier landsat (i.e., LS5 and LS7) data.
    Note, files who's file name contain 'LS8' or 'lc08' are considered to be
    Landsat 8 images. files who's file name contain 'LS9' or 'lc09' are considered
    to be Landsat 8 images.

    :param input_imgs: input list of image file paths.
    :param output_dir: output directory for the VRT images (Note. an absolute path to
                   the input image is used when building the VRT.)
    """
    import rsgislib.tools.filetools

    img_bands = [2, 3, 4, 5, 6, 7]
    out_imgs = []
    for img in input_imgs:
        if ("LS8" in os.path.basename(img)) or ("lc08" in os.path.basename(img)) or ("LS9" in os.path.basename(img)) or ("lc09" in os.path.basename(img)):
            abs_img = os.path.abspath(img)
            basename = rsgislib.tools.filetools.get_file_basename(img)
            out_vrt_img = os.path.join(output_dir, f"{basename}_bsub.vrt")
            rsgislib.imageutils.create_vrt_band_subset(abs_img, img_bands, out_vrt_img)
            out_imgs.append(out_vrt_img)
        else:
            out_imgs.append(img)
    return out_imgs


def create_max_ndvi_composite(
    input_imgs,
    r_band,
    n_band,
    out_ref_img,
    out_comp_img,
    tmp_dir="tmp",
    gdalformat="KEA",
    datatype=None,
    calc_stats=True,
):
    """
    Create an image composite from multiple input images where the pixel brought
    through into the composite is the one with the maximum NDVI.

    :param input_imgs: list of input images for the analysis.
    :param r_band: is the image band within the input images (same for all) for
                   the red band - note band indexing starts at 1.
    :param n_band: is the image band within the input images (same for all) for
                   the nir band - note band indexing starts at 1.
    :param out_ref_img: is the output reference image which details which input
                        image is forming the output image pixel value (Note. this
                        file will always be a KEA file as RAT is used).
    :param out_comp_img: is the output composite image for which gdalformat and
                         datatype define the format and data type.
    :param tmp_dir: is a temp path for intermediate files, if this path doesn't
                    exist is will be created and deleted at runtime.
    :param gdalformat: is the output file format of the out_comp_img, any
                       GDAL compatible format is OK (Defaut is KEA).
    :param datatype: is the data type of the output image (out_comp_img). If
                     None is provided then the data type of the first input image
                     will be used (Default None).
    :param calc_stats: calculate image statics and pyramids (Default=True)

    """
    import rsgislib.tools.utils

    uid_str = rsgislib.tools.utils.uid_generator()

    if len(input_imgs) > 1:
        tmp_present = True
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)
            tmp_present = False

        ref_layers_path = os.path.join(tmp_dir, "ref_lyrs_" + uid_str)

        ref_img_tmp_present = True
        if not os.path.exists(ref_layers_path):
            os.makedirs(ref_layers_path)
            ref_img_tmp_present = False

        if datatype is None:
            datatype = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_imgs[0])

        num_in_lyrs = len(input_imgs)

        numpy.random.seed(5)
        red = numpy.random.randint(1, 255, num_in_lyrs + 1, int)
        red[0] = 0
        numpy.random.seed(2)
        green = numpy.random.randint(1, 255, num_in_lyrs + 1, int)
        green[0] = 0
        numpy.random.seed(9)
        blue = numpy.random.randint(1, 255, num_in_lyrs + 1, int)
        blue[0] = 0
        alpha = numpy.zeros_like(blue)
        alpha[...] = 255
        img_lyrs = numpy.empty(num_in_lyrs + 1, dtype=numpy.dtype("a255"))

        # Generate Comp Ref layers:
        ref_lyrs_lst = []
        idx = 1
        for img in input_imgs:
            print("In Image ({}):\t{}".format(idx, img))
            img_lyrs[idx] = os.path.basename(img)
            base_img_name = os.path.splitext(os.path.basename(img))[0]
            ref_lyr_img = os.path.join(ref_layers_path, base_img_name + "_ndvi.kea")
            rsgislib.imagecalc.calcindices.calc_ndvi(
                img, r_band, n_band, ref_lyr_img, False
            )
            ref_lyrs_lst.append(ref_lyr_img)
            idx = idx + 1
        img_lyrs[0] = ""

        # Create REF Image
        rsgislib.imagecalc.get_img_idx_for_stat(
            ref_lyrs_lst, out_ref_img, "KEA", -999, rsgislib.SUMTYPE_MAX
        )
        if calc_stats:
            # Pop Ref Image with stats
            rsgislib.rastergis.pop_rat_img_stats(out_ref_img, True, True, True)

            # Open the clumps dataset as a gdal dataset
            rat_dataset = osgeo.gdal.Open(out_ref_img, osgeo.gdal.GA_Update)

            # Write colours to RAT
            rios.rat.writeColumn(rat_dataset, "Red", red)
            rios.rat.writeColumn(rat_dataset, "Green", green)
            rios.rat.writeColumn(rat_dataset, "Blue", blue)
            rios.rat.writeColumn(rat_dataset, "Alpha", alpha)
            rios.rat.writeColumn(rat_dataset, "Image", img_lyrs)

            rat_dataset = None

        # Create Composite Image
        rsgislib.imageutils.create_ref_img_composite_img(
            input_imgs, out_comp_img, out_ref_img, gdalformat, datatype, 0.0
        )

        if calc_stats:
            # Calc Stats
            rsgislib.imageutils.pop_img_stats(
                out_comp_img, use_no_data=True, no_data_val=0, calc_pyramids=True
            )

        if not ref_img_tmp_present:
            shutil.rmtree(ref_layers_path, ignore_errors=True)

        if not tmp_present:
            shutil.rmtree(tmp_dir, ignore_errors=True)
    elif len(input_imgs) == 1:
        print("Only 1 Input Image, Just Copying File to output")
        shutil.copy(input_imgs[0], out_comp_img)
    else:
        raise rsgislib.RSGISPyException("At least 2 input images are needed")




def create_max_ndvi_ndwi_composite(
    ref_sp_img,
    input_imgs,
    r_band,
    n_band,
    s_band,
    out_ref_img,
    out_comp_img,
    out_msk_img,
    tmp_path="tmp",
    gdalformat="KEA",
    datatype=None,
    calc_stats=True,
    reproj_method=rsgislib.INTERP_CUBIC,
    use_mode=True,
):
    """
    Create an image composite from multiple input images where the pixel brought
    through into the composite is the one with the maximum NDVI over land and
    NDWI over water. A mask of land and water regions is also produced. The
    reference image is used to define the spatial extent of the output images
    and spatial projection.

    :param ref_sp_img: is a reference image with any number of bands and data
                       type which is used to define the output image extent
                       and projection.
    :param input_imgs: a list of all the input images being used to build the composite.
    :param r_band: is an integer specifying the red band in the input images
                   (starts at 1), used in the NDVI index.
    :param n_band: is an integer specifying the NIR band in the input images
                   (starts at 1), used in the NDVI and NDWI index.
    :param s_band: is an integer specifying the SWIR band in the input images
                   (starts at 1), used in the NDWI index.
    :param out_ref_img: is the output reference image which details which input
                        image is forming the output image pixel value (Note. this
                        file will always be a KEA file as RAT is used).
    :param out_comp_img: is the output composite image for which gdalformat and
                         datatype define the format and data type.
    :param out_msk_img: is the output mask image for regions of water and land
                        where ndvi vs ndwi are used (0=nodata, 1=land, 2=water)
    :param tmp_path: is a temp path for intermediate files, if this path doesn't
                     exist is will be created and deleted at runtime.
    :param gdalformat: is the output file format of the outCompImg, any GDAL
                       compatable format is OK (Defaut is KEA).
    :param datatype: is the data type of the output image (outCompImg). If None
                     is provided then the data type of the first input image
                     will be used (Default None).
    :param calc_stats: calculate image statics and pyramids (Default=True)
    :param reproj_method: specifies the interpolation method (rsgislib.INTERP*) used
                          to reproject the input images which are in a different
                          projection and/or pixel size as the reference image
                          (default: rsgislib.INTERP_CUBIC).
    :param use_mode: True: the land/water masks are combined using the mode
                     and False: the land water masks are combined using median.

    """
    import rsgislib.tools.utils

    uid_str = rsgislib.tools.utils.uid_generator()

    if len(input_imgs) > 1:
        init_in_images = input_imgs
        in_images_overlap = []
        in_images_to_reproj = []
        in_images_sub_to_ref = []
        num_bands = 0
        first = True
        for img in init_in_images:
            if first:
                num_bands = rsgislib.imageutils.get_img_band_count(img)
                first = False
            else:
                band_count = rsgislib.imageutils.get_img_band_count(img)
                if band_count != num_bands:
                    raise rsgislib.RSGISPyException(
                        f"The number of image bands is not "
                        f"consistent (Bands: {num_bands} and {band_count})"
                    )

            same_proj = rsgislib.imageutils.do_gdal_layers_have_same_proj(ref_sp_img, img)
            if rsgislib.imageutils.do_images_overlap(ref_sp_img, img):
                if same_proj:
                    if rsgislib.imageutils.do_img_res_match(ref_sp_img, img):
                        in_images_overlap.append(img)
                    else:
                        in_images_to_reproj.append(img)
                else:
                    in_images_to_reproj.append(img)

        n_imgs = len(in_images_overlap) + len(in_images_to_reproj)
        print(f"There are {n_imgs} images with overlap to create the composite.")
        if n_imgs > 1:
            if datatype is None:
                if len(in_images_overlap) > 0:
                    datatype = rsgislib.imageutils.get_rsgislib_datatype_from_img(
                        in_images_overlap[0]
                    )
                else:
                    datatype = rsgislib.imageutils.get_rsgislib_datatype_from_img(
                        in_images_to_reproj[0]
                    )

            tmp_present = True
            if not os.path.exists(tmp_path):
                os.makedirs(tmp_path)
                tmp_present = False

            subset_img_tmp_present = True
            if len(in_images_overlap) > 0:
                print("Subset images to reference extent.")
                subset_layers_path = os.path.join(tmp_path, f"subset_{uid_str}")
                if not os.path.exists(subset_layers_path):
                    os.makedirs(subset_layers_path)
                    subset_img_tmp_present = False

                for img in in_images_overlap:
                    basename = os.path.splitext(os.path.basename(img))[0]
                    out_img = os.path.join(subset_layers_path, basename + "_subset.kea")
                    rsgislib.imageutils.create_copy_img(
                        ref_sp_img, out_img, num_bands, 0, "KEA", datatype
                    )
                    rsgislib.imageutils.include_imgs_ind_img_intersect(out_img, [img])
                    in_images_sub_to_ref.append(out_img)

            reproj_img_tmp_present = True
            if len(in_images_to_reproj) > 0:
                print("Reproject images to reference extent.")
                reproj_layers_path = os.path.join(tmp_path, "reproj_" + uid_str)
                if not os.path.exists(reproj_layers_path):
                    os.makedirs(reproj_layers_path)
                    reproj_img_tmp_present = False

                for img in in_images_to_reproj:
                    basename = os.path.splitext(os.path.basename(img))[0]
                    out_img = os.path.join(reproj_layers_path, basename + "_reproj.kea")
                    rsgislib.imageutils.resample_img_to_match(
                        ref_sp_img, img, out_img, "KEA", reproj_method, datatype
                    )
                    in_images_sub_to_ref.append(out_img)

            ref_layers_path = os.path.join(tmp_path, f"ref_lyrs_{uid_str}")
            ref_img_tmp_present = True
            if not os.path.exists(ref_layers_path):
                os.makedirs(ref_layers_path)
                ref_img_tmp_present = False

            num_in_lyrs = len(in_images_sub_to_ref)

            numpy.random.seed(5)
            red = numpy.random.randint(1, 255, num_in_lyrs + 1, int)
            red[0] = 0
            numpy.random.seed(2)
            green = numpy.random.randint(1, 255, num_in_lyrs + 1, int)
            green[0] = 0
            numpy.random.seed(9)
            blue = numpy.random.randint(1, 255, num_in_lyrs + 1, int)
            blue[0] = 0
            alpha = numpy.zeros_like(blue)
            alpha[...] = 255
            img_lyrs = numpy.empty(num_in_lyrs + 1, dtype=numpy.dtype("a255"))

            # Generate Comp Ref layers:
            land_water_msk_band_defns = []
            msk_imgs = []
            idx = 1
            for img in in_images_sub_to_ref:
                print("In Image (" + str(idx) + "):\t" + img)
                img_lyrs[idx] = os.path.basename(img)
                base_img_name = os.path.splitext(os.path.basename(img))[0]
                ref_lyr_ndvi_img = os.path.join(ref_layers_path, f"{base_img_name}_ndvi.kea")
                ref_lyr_ndwi_img = os.path.join(ref_layers_path, f"{base_img_name}_ndwi.kea")
                rsgislib.imagecalc.calcindices.calc_ndvi(
                    img, r_band, n_band, ref_lyr_ndvi_img, False
                )
                rsgislib.imagecalc.calcindices.calc_ndwi(
                    img, n_band, s_band, ref_lyr_ndwi_img, False
                )

                ref_lyr_msk_img = os.path.join(
                    ref_layers_path, base_img_name + "_water_land_msk.kea"
                )
                band_defns = []
                band_defns.append(rsgislib.imagecalc.BandDefn("ndvi", ref_lyr_ndvi_img, 1))
                band_defns.append(rsgislib.imagecalc.BandDefn("ndwi", ref_lyr_ndwi_img, 1))
                rsgislib.imagecalc.band_math(
                    ref_lyr_msk_img,
                    "ndvi<-1?0:ndvi>0.3?1:ndwi>0.01?2:1",
                    "KEA",
                    rsgislib.TYPE_8UINT,
                    band_defns,
                )
                msk_imgs.append(ref_lyr_msk_img)
                idx = idx + 1
            img_lyrs[0] = ""

            ref_lyr_msk_stack_img = os.path.join(
                ref_layers_path, uid_str + "_water_land_msk_stack.kea"
            )
            rsgislib.imageutils.stack_img_bands(
                msk_imgs, None, ref_lyr_msk_stack_img, -1, -1, "KEA", rsgislib.TYPE_8UINT
            )
            if use_mode:
                rsgislib.imagecalc.image_pixel_column_summary(
                    ref_lyr_msk_stack_img,
                    out_msk_img,
                    rsgislib.imagecalc.StatsSummary(calc_mode=True),
                    "KEA",
                    rsgislib.TYPE_8UINT,
                    0,
                    True,
                )
            else:
                rsgislib.imagecalc.image_pixel_column_summary(
                    ref_lyr_msk_stack_img,
                    out_msk_img,
                    rsgislib.imagecalc.StatsSummary(calc_median=True),
                    "KEA",
                    rsgislib.TYPE_8UINT,
                    0,
                    True,
                )
            rsgislib.rastergis.pop_rat_img_stats(out_msk_img, True, True, True)

            idx = 1
            ref_lyrs_lst = []
            for img in in_images_sub_to_ref:
                print(f"In Image ({idx}):\t{img}")
                base_img_name = os.path.splitext(os.path.basename(img))[0]
                ref_lyr_ndvi_img = os.path.join(ref_layers_path, f"{base_img_name}_ndvi.kea")
                ref_lyr_ndwi_img = os.path.join(ref_layers_path, f"{base_img_name}_ndwi.kea")
                ref_lyr_lcl_msk_img = os.path.join(
                    ref_layers_path, f"{base_img_name}_water_land_msk.kea"
                )

                ref_lyr_img = os.path.join(ref_layers_path, f"{base_img_name}_ref_hybrid.kea")
                band_defns = []
                band_defns.append(
                    rsgislib.imagecalc.BandDefn("lmsk", ref_lyr_lcl_msk_img, 1)
                )
                band_defns.append(rsgislib.imagecalc.BandDefn("omsk", out_msk_img, 1))
                band_defns.append(rsgislib.imagecalc.BandDefn("ndvi", ref_lyr_ndvi_img, 1))
                band_defns.append(rsgislib.imagecalc.BandDefn("ndwi", ref_lyr_ndwi_img, 1))
                rsgislib.imagecalc.band_math(
                    ref_lyr_img,
                    "lmsk==0?-999:omsk==1?ndvi:omsk==2?ndwi:-999",
                    "KEA",
                    rsgislib.TYPE_32FLOAT,
                    band_defns,
                )
                ref_lyrs_lst.append(ref_lyr_img)
                idx = idx + 1

            # Create REF Image
            rsgislib.imagecalc.get_img_idx_for_stat(
                ref_lyrs_lst, out_ref_img, "KEA", -999, rsgislib.SUMTYPE_MAX
            )
            if calc_stats:
                # Pop Ref Image with stats
                rsgislib.rastergis.pop_rat_img_stats(out_ref_img, True, True, True)

                # Open the clumps dataset as a gdal dataset
                rat_dataset = osgeo.gdal.Open(out_ref_img, osgeo.gdal.GA_Update)
                # Write colours to RAT
                rios.rat.writeColumn(rat_dataset, "Red", red)
                rios.rat.writeColumn(rat_dataset, "Green", green)
                rios.rat.writeColumn(rat_dataset, "Blue", blue)
                rios.rat.writeColumn(rat_dataset, "Alpha", alpha)
                rios.rat.writeColumn(rat_dataset, "Image", img_lyrs)
                rat_dataset = None

            # Create Composite Image
            rsgislib.imageutils.create_ref_img_composite_img(
                in_images_sub_to_ref, out_comp_img, out_ref_img, gdalformat, datatype, 0.0
            )

            if calc_stats:
                # Calc Stats
                rsgislib.imageutils.pop_img_stats(
                    out_comp_img, use_no_data=True, no_data_val=0, calc_pyramids=True
                )

            if not ref_img_tmp_present:
                shutil.rmtree(ref_layers_path, ignore_errors=True)

            if not subset_img_tmp_present:
                shutil.rmtree(subset_layers_path, ignore_errors=True)

            if not reproj_img_tmp_present:
                shutil.rmtree(reproj_layers_path, ignore_errors=True)

            if not tmp_present:
                shutil.rmtree(tmp_path, ignore_errors=True)
        else:
            raise rsgislib.RSGISPyException(f"There were only {n_imgs} images which intersect with the reference image.")

    elif len(input_imgs) == 1:
        print("Only 1 Input Image, Just Copying File to output")
        num_bands = rsgislib.imageutils.get_img_band_count(input_imgs[0])
        if datatype is None:
            datatype = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_imgs[0])
        rsgislib.imageutils.create_copy_img(
            ref_sp_img, out_comp_img, num_bands, 0, gdalformat, datatype
        )
        rsgislib.imageutils.include_imgs_ind_img_intersect(out_comp_img, [input_imgs[0]])
    else:
        raise rsgislib.RSGISPyException("List of input images was empty ")



#!/usr/bin/env python
"""
The tools.geometrytool module contains functions for manipulating and moving files around.
"""

import math
from osgeo import ogr

import rsgislib


def reproj_bbox(bbox, in_osr_proj_obj, out_osr_proj_obj):
    """
    A function to reproject a bounding box.

    :param bbox: input bounding box (MinX, MaxX, MinY, MaxY)
    :param in_osr_proj_obj: an osr.SpatialReference() object representing input projection.
    :param out_osr_proj_obj: an osr.SpatialReference() object representing output projection.

    :return: (MinX, MaxX, MinY, MaxY)

    """
    tlX = bbox[0]
    tlY = bbox[3]
    trX = bbox[1]
    trY = bbox[3]
    brX = bbox[1]
    brY = bbox[2]
    blX = bbox[0]
    blY = bbox[2]

    out_tlX, out_tlY = reproj_point(in_osr_proj_obj, out_osr_proj_obj, tlX, tlY)
    out_trX, out_trY = reproj_point(in_osr_proj_obj, out_osr_proj_obj, trX, trY)
    out_brX, out_brY = reproj_point(in_osr_proj_obj, out_osr_proj_obj, brX, brY)
    out_blX, out_blY = reproj_point(in_osr_proj_obj, out_osr_proj_obj, blX, blY)

    minX = out_tlX
    if out_blX < minX:
        minX = out_blX

    maxX = out_brX
    if out_trX > maxX:
        maxX = out_trX

    minY = out_brY
    if out_blY < minY:
        minY = out_blY

    maxY = out_tlY
    if out_trY > maxY:
        maxY = out_trY

    return [minX, maxX, minY, maxY]


def reproj_bbox_epsg(bbox, in_epsg, out_epsg):
    """
    A function to reproject a bounding box.

    :param bbox: input bounding box (MinX, MaxX, MinY, MaxY)
    :param in_epsg: an EPSG code representing input projection.
    :param out_epsg: an EPSG code representing output projection.
    :return: (MinX, MaxX, MinY, MaxY)

    """
    from osgeo import osr

    in_osr_proj_obj = osr.SpatialReference()
    in_osr_proj_obj.ImportFromEPSG(int(in_epsg))

    out_osr_proj_obj = osr.SpatialReference()
    out_osr_proj_obj.ImportFromEPSG(int(out_epsg))

    out_bbox = reproj_bbox(bbox, in_osr_proj_obj, out_osr_proj_obj)
    return out_bbox


def do_bboxes_intersect(bbox1, bbox2):
    """
    A function which tests whether two bboxes (MinX, MaxX, MinY, MaxY) intersect.

    :param bbox1: The first bounding box (MinX, MaxX, MinY, MaxY)
    :param bbox2: The second bounding box (MinX, MaxX, MinY, MaxY)
    :return: boolean (True they intersect; False they do not intersect)

    """
    x_min = 0
    x_max = 1
    y_min = 2
    y_max = 3
    intersect = (
        (bbox1[x_max] > bbox2[x_min])
        and (bbox2[x_max] > bbox1[x_min])
        and (bbox1[y_max] > bbox2[y_min])
        and (bbox2[y_max] > bbox1[y_min])
    )
    return intersect


def does_bbox_contain(bbox1, bbox2):
    """
    A function which tests whether bbox1 contains bbox2.

    :param bbox1: The first bounding box (MinX, MaxX, MinY, MaxY)
    :param bbox2: The second bounding box (MinX, MaxX, MinY, MaxY)
    :return: boolean (True bbox1 contains bbox2; False bbox1 does not contain bbox2)

    """
    x_min = 0
    x_max = 1
    y_min = 2
    y_max = 3
    contains = (
        (bbox1[x_min] < bbox2[x_min])
        and (bbox1[x_max] > bbox2[x_max])
        and (bbox1[y_min] < bbox2[y_min])
        and (bbox1[y_max] > bbox2[y_max])
    )
    return contains


def calc_bbox_area(bbox):
    """
    Calculate the area of the bbox.

    :param bbox: bounding box (MinX, MaxX, MinY, MaxY)
    :return: area in projection of the bbox.

    """
    # width x height
    return (bbox[1] - bbox[0]) * (bbox[3] - bbox[2])


def bbox_equal(bbox1, bbox2):
    """
    A function which tests whether two bboxes (xMin, xMax, yMin, yMax) are equal.

    :param bbox1: is a bbox (xMin, xMax, yMin, yMax)
    :param bbox2: is a bbox (xMin, xMax, yMin, yMax)
    :return: boolean

    """
    bbox_eql = False
    if (
        (bbox1[0] == bbox2[0])
        and (bbox1[1] == bbox2[1])
        and (bbox1[2] == bbox2[2])
        and (bbox1[3] == bbox2[3])
    ):
        bbox_eql = True
    return bbox_eql


def bbox_intersection(bbox1, bbox2):
    """
    A function which calculates the intersection of the two bboxes (xMin, xMax, yMin, yMax).

    :param bbox1: is a bbox (xMin, xMax, yMin, yMax)
    :param bbox2: is a bbox (xMin, xMax, yMin, yMax)
    :return: bbox (xMin, xMax, yMin, yMax)

    """
    if not do_bboxes_intersect(bbox1, bbox2):
        raise rsgislib.RSGISPyException("Bounding Boxes do not intersect.")

    xMinOverlap = bbox1[0]
    xMaxOverlap = bbox1[1]
    yMinOverlap = bbox1[2]
    yMaxOverlap = bbox1[3]

    if bbox2[0] > xMinOverlap:
        xMinOverlap = bbox2[0]

    if bbox2[1] < xMaxOverlap:
        xMaxOverlap = bbox2[1]

    if bbox2[2] > yMinOverlap:
        yMinOverlap = bbox2[2]

    if bbox2[3] < yMaxOverlap:
        yMaxOverlap = bbox2[3]

    return [xMinOverlap, xMaxOverlap, yMinOverlap, yMaxOverlap]


def bboxes_intersection(bboxes):
    """
    A function to find the intersection between a list of
    bboxes.

    :param bboxes: a list of bboxes [(xMin, xMax, yMin, yMax)]
    :return: bbox (xMin, xMax, yMin, yMax)

    """
    if len(bboxes) == 1:
        return bboxes[0]
    elif len(bboxes) == 2:
        return bbox_intersection(bboxes[0], bboxes[1])

    inter_bbox = bboxes[0]
    for bbox in bboxes[1:]:
        inter_bbox = bbox_intersection(inter_bbox, bbox)
    return inter_bbox


def buffer_bbox(bbox, buf):
    """
    Buffer the input BBOX by a set amount.

    :param bbox: the bounding box (MinX, MaxX, MinY, MaxY)
    :param buf: the amount of buffer by
    :return: the buffered bbox (MinX, MaxX, MinY, MaxY)

    """
    out_bbox = [0, 0, 0, 0]
    out_bbox[0] = bbox[0] - buf
    out_bbox[1] = bbox[1] + buf
    out_bbox[2] = bbox[2] - buf
    out_bbox[3] = bbox[3] + buf
    return out_bbox


def find_bbox_union(bboxes):
    """
    A function which finds the union of all the bboxes inputted.

    :param bboxes: a list of bboxes [(xMin, xMax, yMin, yMax), (xMin, xMax, yMin, yMax)]
    :return: bbox (xMin, xMax, yMin, yMax)

    """
    if len(bboxes) == 1:
        out_bbox = list(bboxes[0])
    elif len(bboxes) > 1:
        out_bbox = list(bboxes[0])
        for bbox in bboxes:
            if bbox[0] < out_bbox[0]:
                out_bbox[0] = bbox[0]
            if bbox[1] > out_bbox[1]:
                out_bbox[1] = bbox[1]
            if bbox[2] < out_bbox[2]:
                out_bbox[2] = bbox[2]
            if bbox[3] > out_bbox[3]:
                out_bbox[3] = bbox[3]
    else:
        out_bbox = None
    return out_bbox


def unwrap_wgs84_bbox(bbox):
    """
    A function which unwraps a bbox if it projected in WGS84 and over the 180/-180 boundary.

    :param bbox: the bounding box (MinX, MaxX, MinY, MaxY)
    :return: list of bounding boxes [(MinX, MaxX, MinY, MaxY), (MinX, MaxX, MinY, MaxY)...]

    """
    bboxes = []
    if bbox[1] < bbox[0]:
        bbox1 = [-180.0, bbox[1], bbox[2], bbox[3]]
        bboxes.append(bbox1)
        bbox2 = [bbox[0], 180.0, bbox[2], bbox[3]]
        bboxes.append(bbox2)
    else:
        bboxes.append(bbox)
    return bboxes


def find_common_extent_on_grid(base_extent, base_grid, other_extent, full_contain=True):
    """
    A function which calculates the common extent between two extents but defines output on
    grid with defined resolutions. Useful for finding common extent on a particular image grid.

    :param base_extent: is a bbox (xMin, xMax, yMin, yMax) providing the base for the grid on which output will be defined.
    :param base_grid: the size of the (square) grid on which output will be defined.
    :param other_extent: is a bbox (xMin, xMax, yMin, yMax) to be intersected with the base_extent.
    :param full_contain: is a boolean. True: moving output onto grid will increase size of bbox (i.e., intersection fully contained)
                                      False: move output onto grid will decrease size of bbox (i.e., bbox fully contained within intesection)

    :return: bbox (xMin, xMax, yMin, yMax)

    """
    xMinOverlap = base_extent[0]
    xMaxOverlap = base_extent[1]
    yMinOverlap = base_extent[2]
    yMaxOverlap = base_extent[3]

    if other_extent[0] > xMinOverlap:
        if full_contain:
            diff = math.floor((other_extent[0] - xMinOverlap) / base_grid) * base_grid
        else:
            diff = math.ceil((other_extent[0] - xMinOverlap) / base_grid) * base_grid
        xMinOverlap = xMinOverlap + diff

    if other_extent[1] < xMaxOverlap:
        if full_contain:
            diff = math.floor((xMaxOverlap - other_extent[1]) / base_grid) * base_grid
        else:
            diff = math.ceil((xMaxOverlap - other_extent[1]) / base_grid) * base_grid
        xMaxOverlap = xMaxOverlap - diff

    if other_extent[2] > yMinOverlap:
        if full_contain:
            diff = (
                math.floor(abs(other_extent[2] - yMinOverlap) / base_grid) * base_grid
            )
        else:
            diff = math.ceil(abs(other_extent[2] - yMinOverlap) / base_grid) * base_grid
        yMinOverlap = yMinOverlap + diff

    if other_extent[3] < yMaxOverlap:
        if full_contain:
            diff = (
                math.floor(abs(yMaxOverlap - other_extent[3]) / base_grid) * base_grid
            )
        else:
            diff = math.ceil(abs(yMaxOverlap - other_extent[3]) / base_grid) * base_grid
        yMaxOverlap = yMaxOverlap - diff

    return [xMinOverlap, xMaxOverlap, yMinOverlap, yMaxOverlap]


def find_extent_on_grid(base_extent, base_grid, full_contain=True):
    """
    A function which calculates the extent but defined on a grid with defined resolution.
    Useful for finding extent on a particular image grid.

    :param base_extent: is a bbox (xMin, xMax, yMin, yMax) providing the base for the grid on which output will be defined.
    :param base_grid: the size of the (square) grid on which output will be defined.
    :param full_contain: is a boolean. True: moving output onto grid will increase size of bbox (i.e., intersection fully contained)
                                      False: move output onto grid will decrease size of bbox (i.e., bbox fully contained within intesection)

    :return: bbox (xMin, xMax, yMin, yMax)

    """
    xMin = base_extent[0]
    xMax = base_extent[1]
    yMin = base_extent[2]
    yMax = base_extent[3]

    diffX = xMax - xMin
    diffY = abs(yMax - yMin)

    nPxlX = 0.0
    nPxlY = 0.0
    if full_contain:
        nPxlX = math.ceil(diffX / base_grid)
        nPxlY = math.ceil(diffY / base_grid)
    else:
        nPxlX = math.floor(diffX / base_grid)
        nPxlY = math.floor(diffY / base_grid)

    xMaxOut = xMin + (nPxlX * base_grid)
    yMinOut = yMax - (nPxlY * base_grid)

    return [xMin, xMaxOut, yMinOut, yMax]


def find_extent_on_whole_num_grid(
    base_extent, base_grid, full_contain=True, round_vals=None
):
    """
    A function which calculates the extent but defined on a grid with defined resolution.
    Useful for finding extent on a particular image grid.

    :param base_extent: is a bbox (xMin, xMax, yMin, yMax) providing the base for the grid on which output will be defined.
    :param base_grid: the size of the (square) grid on which output will be defined.
    :param full_contain: is a boolean. True: moving output onto grid will increase size of bbox (i.e., intersection fully contained)
                                      False: move output onto grid will decrease size of bbox (i.e., bbox fully contained within intesection)
    :param round_vals: specify whether outputted values should be rounded. None for no rounding (default) or integer for number of
                       significant figures to round to.

    :return: bbox (xMin, xMax, yMin, yMax)

    """
    xMin = base_extent[0]
    xMax = base_extent[1]
    yMin = base_extent[2]
    yMax = base_extent[3]

    nPxlXMin = math.floor(xMin / base_grid)
    nPxlYMin = math.floor(yMin / base_grid)

    xMinOut = nPxlXMin * base_grid
    yMinOut = nPxlYMin * base_grid

    diffX = xMax - xMinOut
    diffY = abs(yMax - yMinOut)

    nPxlX = 0.0
    nPxlY = 0.0
    if full_contain:
        nPxlX = math.ceil(diffX / base_grid)
        nPxlY = math.ceil(diffY / base_grid)
    else:
        nPxlX = math.floor(diffX / base_grid)
        nPxlY = math.floor(diffY / base_grid)

    xMaxOut = xMinOut + (nPxlX * base_grid)
    yMaxOut = yMinOut + (nPxlY * base_grid)

    if round_vals is None:
        out_bbox = [xMinOut, xMaxOut, yMinOut, yMaxOut]
    else:
        out_bbox = [
            round(xMinOut, round_vals),
            round(xMaxOut, round_vals),
            round(yMinOut, round_vals),
            round(yMaxOut, round_vals),
        ]
    return out_bbox


def get_bbox_grid(bbox, x_size, y_size):
    """
    Create a grid with size x_size, y_size for the area represented by bbox.

    :param bbox: a bounding box within which the grid will be created (xMin, xMax, yMin, yMax)
    :param x_size: Output grid size in X axis (same unit as bbox).
    :param y_size: Output grid size in Y axis (same unit as bbox).

    :return: list of bounding boxes (xMin, xMax, yMin, yMax)

    """
    width = bbox[1] - bbox[0]
    height = bbox[3] - bbox[2]

    n_tiles_x = math.floor(width / x_size)
    n_tiles_y = math.floor(height / y_size)

    if (n_tiles_x > 10000) or (n_tiles_y > 10000):
        print(
            "WARNING: did you mean to product so many tiles (X: {}, Y: {}) "
            "might want to check your units".format(n_tiles_x, n_tiles_y)
        )

    full_tile_width = n_tiles_x * x_size
    full_tile_height = n_tiles_y * y_size

    x_remain = width - full_tile_width
    if x_remain < 0.000001:
        x_remain = 0.0
    y_remain = height - full_tile_height
    if y_remain < 0.000001:
        y_remain = 0.0

    c_min_y = bbox[2]
    c_max_y = c_min_y + y_size

    bboxs = list()
    for ny in range(n_tiles_y):
        c_min_x = bbox[0]
        c_max_x = c_min_x + x_size
        for nx in range(n_tiles_x):
            bboxs.append([c_min_x, c_max_x, c_min_y, c_max_y])
            c_min_x = c_max_x
            c_max_x = c_max_x + x_size
        if x_remain > 0:
            c_max_x = c_min_x + x_remain
            bboxs.append([c_min_x, c_max_x, c_min_y, c_max_y])
        c_min_y = c_max_y
        c_max_y = c_max_y + y_size
    if y_remain > 0:
        c_max_y = c_min_y + y_remain
        c_min_x = bbox[0]
        c_max_x = c_min_x + x_size
        for nx in range(n_tiles_x):
            bboxs.append([c_min_x, c_max_x, c_min_y, c_max_y])
            c_min_x = c_max_x
            c_max_x = c_max_x + x_size
        if x_remain > 0:
            c_max_x = c_min_x + x_remain
            bboxs.append([c_min_x, c_max_x, c_min_y, c_max_y])

    return bboxs


def reproj_point(in_osr_proj_obj, out_osr_proj_obj, x, y):
    """
    Reproject a point from 'in_osr_proj_obj' to 'out_osr_proj_obj' where they are gdal
    osgeo.osr.SpatialReference objects.

    :return: x, y.

    """
    if in_osr_proj_obj.EPSGTreatsAsLatLong():
        wktPt = "POINT(%s %s)" % (y, x)
    else:
        wktPt = "POINT(%s %s)" % (x, y)
    point = ogr.CreateGeometryFromWkt(wktPt)
    point.AssignSpatialReference(in_osr_proj_obj)
    point.TransformTo(out_osr_proj_obj)
    if out_osr_proj_obj.EPSGTreatsAsLatLong():
        outX = point.GetY()
        outY = point.GetX()
    else:
        outX = point.GetX()
        outY = point.GetY()
    return outX, outY



#! /usr/bin/env python
############################################################################
#  classratutils.py
#
#  Copyright 2015 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 28/11/2015
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################
import os

from osgeo import gdal
import numpy
from rios import rat

import rsgislib


def populate_clumps_with_class_training(
    clumps_img: str,
    classes_info: list,
    tmp_dir: str,
    classes_int_col: str,
    classes_name_col: str,
    rat_band: int = 1,
):
    """
    A function to populate a clumps file with training from a series of
    vector layers (1 per class)

    :param clumps_img: input clumps file.
    :param classes_info: A list of rsgislb.classification.ClassVecSamplesInfoObj
                         objects. Note, the file_h5 variable is not needed in
                         this function.
    :param tmp_dir: File path (which needs to exist) where files can
                    temporally be written.
    :param classes_int_col: Output column name for integer values representing
                            each class.
    :param classes_name_col: Output column name for string class names.
    :param rat_band: The band within the input image the RAT is associated with.

    """
    import rsgislib.imageutils
    import rsgislib.rastergis
    import rsgislib.vectorutils.createrasters
    import rsgislib.tools.utils
    import rsgislib.tools.filetools
    import shutil

    createdDIR = False
    if not os.path.isdir(tmp_dir):
        os.makedirs(tmp_dir)
        createdDIR = True

    uid = rsgislib.tools.utils.uid_generator(10)

    classLayerSeq = list()
    tmpClassImgLayers = list()
    classNamesDict = dict()

    for cls_info in classes_info:

        tmp_cls_img = os.path.join(
            tmp_dir, "{}_{}.kea".format(cls_info.class_name, uid)
        )
        rsgislib.vectorutils.createrasters.rasterise_vec_lyr(
            cls_info.vec_file,
            cls_info.vec_lyr,
            clumps_img,
            tmp_cls_img,
            gdalformat="KEA",
            burn_val=cls_info.id,
            datatype=rsgislib.TYPE_16UINT,
        )
        tmpClassImgLayers.append(tmp_cls_img)
        classNamesDict[cls_info.id] = cls_info.class_name

    combinedClassesImage = os.path.join(tmp_dir, "CombinedClasses_{}.kea".format(uid))
    rsgislib.imageutils.combine_imgs_to_band(
        tmpClassImgLayers, combinedClassesImage, "KEA", rsgislib.TYPE_8UINT, 0.0
    )

    rsgislib.rastergis.populate_rat_with_mode(
        input_img=combinedClassesImage,
        clumps_img=clumps_img,
        out_cols_name=classes_int_col,
        use_no_data=False,
        no_data_val=0,
        out_no_data=False,
        mode_band=1,
        rat_band=rat_band,
    )
    rsgislib.rastergis.define_class_names(
        clumps_img, classes_int_col, classes_name_col, classNamesDict
    )

    for file in tmpClassImgLayers:
        rsgislib.tools.filetools.delete_file_with_basename(file)
    rsgislib.tools.filetools.delete_file_with_basename(combinedClassesImage)

    if createdDIR:
        shutil.rmtree(tmp_dir)


def extract_rat_col_data(
    clumps_img: str,
    cols: list,
    sel_col: str,
    sel_col_val: str,
    out_h5_file: str,
    datatype: int = None,
    rat_band: int = 1,
):
    """
    A function which extracts column values to be used as training, testing, validation
    sets for building a classifier. The data will be saved within a HDF5 file. Note,
    this function reads each whole column into memory and then subsets it so for very
    large RATs this might cause problems.

    :param clumps_img: The inputted clumps file with the associated RAT.
    :param cols: a list of the columns to be exported. Note, they will be stored in the
                 HDF5 file in the order specified here and therefore this order is
                 important and needs to be maintained when this is used going forward
                 (e.g., when applying a classifier trained using this data.
    :param sel_col: The column within the RAT specifying which rows will be exported.
    :param sel_col_val: The value in the sel_col which indicated which rows are to be
                        exported.
    :param out_h5_file: The output HDF5 file the data will be exported to.
    :param datatype: is the data type used for the output HDF5 file (e.g.,
                     rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.
    :param rat_band: The band within the input image the RAT is associated with.

    """
    import h5py

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT
    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)

    rat_dataset = gdal.Open(clumps_img, gdal.GA_ReadOnly)
    sel_col_vals = rat.readColumn(rat_dataset, sel_col, bandNumber=rat_band)
    sel_col_sub_vals = sel_col_vals[sel_col_vals == sel_col_val]

    n_feats = sel_col_sub_vals.shape[0]
    print("n feats: {}".format(n_feats))

    n_cols = len(cols)
    print("n cols: {}".format(n_cols))

    out_vars_arr = numpy.zeros((n_feats, n_cols), dtype=h5_dtype)

    for i, col in enumerate(cols):
        col_data = rat.readColumn(rat_dataset, col, bandNumber=rat_band)
        col_data = col_data[sel_col_vals == sel_col_val]
        out_vars_arr[..., i] = col_data

    rat_dataset = None

    chunk_len = 1000
    if n_feats < chunk_len:
        chunk_len = n_feats

    fH5Out = h5py.File(out_h5_file, "w")
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    dataGrp.create_dataset(
        "DATA",
        data=out_vars_arr,
        chunks=(chunk_len, n_cols),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S255")
    describDS[0] = "Extracted from: ".format(os.path.basename(clumps_img)).encode()
    fH5Out.close()



import os
import pytest

SKLEARN_NOT_AVAIL = False
try:
    import sklearn
except ImportError:
    SKLEARN_NOT_AVAIL = True

MATPLOTLIB_NOT_AVAIL = False
try:
    import matplotlib.pyplot
except ImportError:
    MATPLOTLIB_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
CLASS_ACC_DATA_DIR = os.path.join(DATA_DIR, "classification", "accuracy")


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_calc_acc_ptonly_metrics_vecsamples(tmp_path):
    import rsgislib.classification.classaccuracymetrics

    vec_file = os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_1.geojson")
    vec_lyr = "gmw_acc_roi_1_cls_acc_pts_1"

    out_json_file = os.path.join(tmp_path, "out_acc_stats.json")
    out_csv_file = os.path.join(tmp_path, "out_acc_stats.csv")

    ref_col = "ref_cls"
    cls_col = "gmw_v25_cls"

    rsgislib.classification.classaccuracymetrics.calc_acc_ptonly_metrics_vecsamples(
        vec_file, vec_lyr, ref_col, cls_col, out_json_file, out_csv_file
    )

    assert os.path.exists(out_json_file) and os.path.exists(out_csv_file)


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_calc_acc_ptonly_metrics_vecsamples_bootstrap_conf_interval(tmp_path):
    import rsgislib.classification.classaccuracymetrics

    vec_file = os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_1.geojson")
    vec_lyr = "gmw_acc_roi_1_cls_acc_pts_1"

    out_json_file = os.path.join(tmp_path, "out_acc_stats.json")

    ref_col = "ref_cls"
    cls_col = "gmw_v25_cls"

    rsgislib.classification.classaccuracymetrics.calc_acc_ptonly_metrics_vecsamples_bootstrap_conf_interval(
        vec_file,
        vec_lyr,
        ref_col,
        cls_col,
        out_json_file,
        sample_frac=0.2,
        sample_n_smps=100,
        bootstrap_n=100,
    )

    assert os.path.exists(out_json_file)


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_calc_acc_ptonly_metrics_vecsamples_f1_conf_inter_sets(tmp_path):
    import rsgislib.classification.classaccuracymetrics
    import rsgislib.tools.filetools

    vec_files = [
        os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_1.geojson"),
        os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_2.geojson"),
        os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_3.geojson"),
        os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_4.geojson"),
        os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_5.geojson"),
    ]

    vec_lyrs = list()
    for vec_file in vec_files:
        vec_lyrs.append(rsgislib.tools.filetools.get_file_basename(vec_file))

    ref_col = "ref_cls"
    cls_col = "gmw_v25_cls"

    out_plot_file = os.path.join(tmp_path, "out_plot.png")

    (
        conf_thres_met,
        conf_thres_met_idx,
        f1_scores,
        f1_scr_intervals_rgn,
    ) = rsgislib.classification.classaccuracymetrics.calc_acc_ptonly_metrics_vecsamples_f1_conf_inter_sets(
        vec_files,
        vec_lyrs,
        ref_col,
        cls_col,
        tmp_path,
        conf_inter=95,
        conf_thres=0.05,
        out_plot_file=out_plot_file,
        sample_frac=0.5,
        sample_n_smps=100,
        bootstrap_n=100,
    )

    assert os.path.exists(out_plot_file)


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_summarise_multi_acc_ptonly_metrics(tmp_path):
    import rsgislib.classification.classaccuracymetrics
    import rsgislib.tools.filetools

    ref_col = "ref_cls"
    cls_col = "gmw_v25_cls"

    vec_files = [
        os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_1.geojson"),
        os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_2.geojson"),
        os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_3.geojson"),
        os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_4.geojson"),
        os.path.join(CLASS_ACC_DATA_DIR, "gmw_acc_roi_1_cls_acc_pts_5.geojson"),
    ]

    acc_json_files = list()

    for vec_file in vec_files:
        vec_lyr = rsgislib.tools.filetools.get_file_basename(vec_file)
        out_json_file = os.path.join(tmp_path, "{}_stats.json".format(vec_lyr))
        rsgislib.classification.classaccuracymetrics.calc_acc_ptonly_metrics_vecsamples(
            vec_file, vec_lyr, ref_col, cls_col, out_json_file
        )
        acc_json_files.append(out_json_file)

    out_acc_json_sum_file = os.path.join(tmp_path, "out_acc_stats.json")
    rsgislib.classification.classaccuracymetrics.summarise_multi_acc_ptonly_metrics(
        acc_json_files, out_acc_json_sum_file
    )

    assert os.path.exists(out_json_file)


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_calc_acc_metrics_vecsamples(tmp_path):
    import rsgislib.classification.classaccuracymetrics

    vec_file = os.path.join(CLASS_ACC_DATA_DIR, "cls_acc_assessment_pts_ref.geojson")
    vec_lyr = "cls_acc_assessment_pts_ref"

    in_cls_img = os.path.join(CLASS_ACC_DATA_DIR, "cls_rf_refl.kea")

    out_json_file = os.path.join(tmp_path, "out_acc_stats.json")
    out_csv_file = os.path.join(tmp_path, "out_acc_stats.csv")

    rsgislib.classification.classaccuracymetrics.calc_acc_metrics_vecsamples(
        vec_file=vec_file,
        vec_lyr=vec_lyr,
        ref_col="ref_pts",
        cls_col="rf_rl_cls",
        cls_img=in_cls_img,
        img_cls_name_col="class_names",
        img_hist_col="Histogram",
        out_json_file=out_json_file,
        out_csv_file=out_csv_file,
    )

    assert os.path.exists(out_json_file) and os.path.exists(out_csv_file)


def test_create_norm_modelled_err_matrix():
    from rsgislib.classification.classaccuracymetrics import (
        create_norm_modelled_err_matrix,
    )

    cls_areas = [40, 40, 10, 10]

    rel_err_mtx = [
        [75.0, 5.0, 5.0, 15.0],
        [5.0, 75.0, 2.0, 18.0],
        [20.0, 2.5, 75.0, 2.5],
        [2.5, 20.0, 2.5, 75.0],
    ]

    err_mtx = create_norm_modelled_err_matrix(cls_areas, rel_err_mtx)


def test_create_modelled_acc_pts():
    import rsgislib.classification.classaccuracymetrics

    cls_lst = [
        "Mangroves",
        "Not Mangroves",
        "Mangroves to Not Mangroves",
        "Not Mangroves to Mangroves",
    ]

    err_mtx_unit_area = [
        [0.378, 0.02, 0.002, 0.0],
        [0.02, 0.478, 0.0, 0.002],
        [0.02, 0.0, 0.03, 0.0],
        [0.0, 0.02, 0.0, 0.03],
    ]

    (
        ref_samples,
        pred_samples,
    ) = rsgislib.classification.classaccuracymetrics.create_modelled_acc_pts(
        err_mtx_unit_area, cls_lst, 1600
    )


@pytest.mark.skipif(
    (MATPLOTLIB_NOT_AVAIL or SKLEARN_NOT_AVAIL),
    reason="matplotlib or scikit-learn dependency not available",
)
def test_calc_sampled_acc_metrics(tmp_path):
    from rsgislib.classification import classaccuracymetrics

    cls_lst = [
        "Mangroves",
        "Not Mangroves",
        "Mangroves to Not Mangroves",
        "Not Mangroves to Mangroves",
    ]

    err_mtx_unit_area = [
        [0.378, 0.02, 0.002, 0.0],
        [0.02, 0.478, 0.0, 0.002],
        [0.02, 0.0, 0.03, 0.0],
        [0.0, 0.02, 0.0, 0.03],
    ]

    (
        ref_samples,
        pred_samples,
    ) = classaccuracymetrics.create_modelled_acc_pts(err_mtx_unit_area, cls_lst, 2500)

    smpls_lst = [500, 1000, 1500, 2000]

    out_metrics_file = os.path.join(tmp_path, "out_stats_file.json")
    out_usr_metrics_plot = os.path.join(tmp_path, "usr_metrics_plot.png")
    out_prod_metrics_plot = os.path.join(tmp_path, "prod_metrics_plot.png")
    out_ref_usr_plot = os.path.join(tmp_path, "usr_ref_metrics_plot.png")
    out_ref_prod_plot = os.path.join(tmp_path, "prod_ref_metrics_plot.png")

    classaccuracymetrics.calc_sampled_acc_metrics(
        ref_samples,
        pred_samples,
        cls_lst,
        smpls_lst,
        out_metrics_file,
        n_repeats=10,
        out_usr_metrics_plot=out_usr_metrics_plot,
        out_prod_metrics_plot=out_prod_metrics_plot,
        out_ref_usr_plot=out_ref_usr_plot,
        out_ref_prod_plot=out_ref_prod_plot,
    )



import os
import pytest
from shutil import copy2

H5PY_NOT_AVAIL = False
try:
    import h5py
except ImportError:
    H5PY_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
CLASSIFICATION_DATA_DIR = os.path.join(DATA_DIR, "classification")


def test_populate_clumps_with_class_training(tmp_path):
    import rsgislib.classification
    import rsgislib.classification.classratutils
    import rsgislib.rastergis
    import numpy

    cls_vec_smpls = os.path.join(CLASSIFICATION_DATA_DIR, "cls_poly_smpls.gpkg")

    clumps_img_ref = os.path.join(
        CLASSIFICATION_DATA_DIR, "sen2_20210527_aber_clumps_s2means.kea"
    )
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps_s2means.kea")
    copy2(clumps_img_ref, clumps_img)

    class_vec_sample_info = []
    class_vec_sample_info.append(
        rsgislib.classification.ClassVecSamplesInfoObj(
            id=1,
            class_name="Forest",
            vec_file=cls_vec_smpls,
            vec_lyr="cls_forest_smpls",
            file_h5="cls_forest_smpls.h5",
        )
    )
    class_vec_sample_info.append(
        rsgislib.classification.ClassVecSamplesInfoObj(
            id=2,
            class_name="Grass",
            vec_file=cls_vec_smpls,
            vec_lyr="cls_grass_smpls",
            file_h5="cls_grass_smpls.h5",
        )
    )
    class_vec_sample_info.append(
        rsgislib.classification.ClassVecSamplesInfoObj(
            id=3,
            class_name="Urban",
            vec_file=cls_vec_smpls,
            vec_lyr="cls_urban_smpls",
            file_h5="cls_urban_smpls.h5",
        )
    )
    class_vec_sample_info.append(
        rsgislib.classification.ClassVecSamplesInfoObj(
            id=4,
            class_name="Water",
            vec_file=cls_vec_smpls,
            vec_lyr="cls_water_smpls",
            file_h5="cls_water_smpls.h5",
        )
    )

    rsgislib.classification.classratutils.populate_clumps_with_class_training(
        clumps_img, class_vec_sample_info, "tmp", "cls_int", "cls_name"
    )

    clumps_cls_img = os.path.join(
        CLASSIFICATION_DATA_DIR, "sen2_20210527_aber_clumps_s2means_cls.kea"
    )

    vars_to_test = ["cls_int", "cls_name"]
    vars_eq_vals = True
    for var in vars_to_test:
        print("Testing: {}".format(var))

        ref_vals = rsgislib.rastergis.get_column_data(clumps_cls_img, var)
        calcd_vals = rsgislib.rastergis.get_column_data(clumps_img, var)
        if not numpy.array_equal(ref_vals, calcd_vals):
            vars_eq_vals = False
            break

    assert vars_eq_vals


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_extract_rat_col_data(tmp_path):
    import rsgislib.classification.classratutils

    clumps_img = os.path.join(
        CLASSIFICATION_DATA_DIR, "sen2_20210527_aber_clumps_s2means_cls.kea"
    )

    cols = [
        "b1Mean",
        "b2Mean",
        "b3Mean",
        "b4Mean",
        "b5Mean",
        "b6Mean",
        "b7Mean",
        "b8Mean",
        "b9Mean",
        "b10Mean",
    ]
    out_h5_file = os.path.join(tmp_path, "forest_smpls.h5")
    rsgislib.classification.classratutils.extract_rat_col_data(
        clumps_img, cols, sel_col="cls_int", sel_col_val=1, out_h5_file=out_h5_file
    )

    assert os.path.exists(out_h5_file)



import os
import pytest
from shutil import copy2

H5PY_NOT_AVAIL = False
try:
    import h5py
except ImportError:
    H5PY_NOT_AVAIL = True

PLOTLY_NOT_AVAIL = False
try:
    import plotly
except ImportError:
    PLOTLY_NOT_AVAIL = True

KALEIDO_NOT_AVAIL = False
try:
    import kaleido
except ImportError:
    KALEIDO_NOT_AVAIL = True

GEOPANDAS_NOT_AVAIL = False
try:
    import geopandas
except ImportError:
    GEOPANDAS_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
ZONALSTATS_DATA_DIR = os.path.join(DATA_DIR, "zonalstats")
CLASSIFICATION_DATA_DIR = os.path.join(DATA_DIR, "classification")


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_get_num_samples_pts():
    import rsgislib.classification

    in_h5_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_vals.h5")

    n_smpls = rsgislib.classification.get_num_samples(in_h5_file)

    assert n_smpls == 703


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_get_num_samples_chips():
    import rsgislib.classification

    in_h5_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_chip_vals.h5"
    )

    n_smpls = rsgislib.classification.get_num_samples(in_h5_file)

    assert n_smpls == 6


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_get_num_samples_refchips():
    import rsgislib.classification

    in_h5_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_refchip_vals.h5"
    )

    n_smpls = rsgislib.classification.get_num_samples(in_h5_file)

    assert n_smpls == 6


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_split_sample_train_valid_test(tmp_path):
    import rsgislib.classification

    in_h5_file = os.path.join(ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_vals.h5")

    train_h5_file = os.path.join(tmp_path, "out_train_data.h5")
    valid_h5_file = os.path.join(tmp_path, "out_valid_data.h5")
    test_h5_file = os.path.join(tmp_path, "out_test_data.h5")

    rsgislib.classification.split_sample_train_valid_test(
        in_h5_file,
        train_h5_file,
        valid_h5_file,
        test_h5_file,
        100,
        100,
        train_sample=500,
        rnd_seed=42,
        datatype=rsgislib.TYPE_16INT,
    )

    assert (
        os.path.exists(train_h5_file)
        and os.path.exists(valid_h5_file)
        and os.path.exists(test_h5_file)
    )


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_create_train_valid_test_sets(tmp_path):
    import rsgislib.classification

    cls_in_info = dict()
    cls_in_info["Forest"] = rsgislib.classification.ClassSimpleInfoObj(
        id=1,
        file_h5=os.path.join(CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal.h5"),
        red=120,
        green=120,
        blue=120,
    )
    cls_in_info["Grass"] = rsgislib.classification.ClassSimpleInfoObj(
        id=2,
        file_h5=os.path.join(CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal.h5"),
        red=120,
        green=120,
        blue=120,
    )
    cls_in_info["Urban"] = rsgislib.classification.ClassSimpleInfoObj(
        id=3,
        file_h5=os.path.join(CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal.h5"),
        red=120,
        green=120,
        blue=120,
    )
    cls_in_info["Water"] = rsgislib.classification.ClassSimpleInfoObj(
        id=4,
        file_h5=os.path.join(CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal.h5"),
        red=120,
        green=120,
        blue=120,
    )

    cls_out_info = dict()
    cls_out_info["Forest"] = rsgislib.classification.ClassInfoObj(
        id=0,
        out_id=1,
        train_file_h5=os.path.join(tmp_path, "cls_forest_smpls_bal_train.h5"),
        test_file_h5=os.path.join(tmp_path, "cls_forest_smpls_bal_test.h5"),
        valid_file_h5=os.path.join(tmp_path, "cls_forest_smpls_bal_valid.h5"),
        red=120,
        green=120,
        blue=120,
    )
    cls_out_info["Grass"] = rsgislib.classification.ClassInfoObj(
        id=1,
        out_id=2,
        train_file_h5=os.path.join(tmp_path, "cls_grass_smpls_bal_train.h5"),
        test_file_h5=os.path.join(tmp_path, "cls_grass_smpls_bal_test.h5"),
        valid_file_h5=os.path.join(tmp_path, "cls_grass_smpls_bal_valid.h5"),
        red=120,
        green=120,
        blue=120,
    )
    cls_out_info["Urban"] = rsgislib.classification.ClassInfoObj(
        id=2,
        out_id=3,
        train_file_h5=os.path.join(tmp_path, "cls_urban_smpls_bal_train.h5"),
        test_file_h5=os.path.join(tmp_path, "cls_urban_smpls_bal_test.h5"),
        valid_file_h5=os.path.join(tmp_path, "cls_urban_smpls_bal_valid.h5"),
        red=120,
        green=120,
        blue=120,
    )
    cls_out_info["Water"] = rsgislib.classification.ClassInfoObj(
        id=3,
        out_id=4,
        train_file_h5=os.path.join(tmp_path, "cls_water_smpls_bal_train.h5"),
        test_file_h5=os.path.join(tmp_path, "cls_water_smpls_bal_test.h5"),
        valid_file_h5=os.path.join(tmp_path, "cls_water_smpls_bal_valid.h5"),
        red=120,
        green=120,
        blue=120,
    )

    rsgislib.classification.create_train_valid_test_sets(
        cls_in_info, cls_out_info, 500, 500, train_sample=None
    )

    assert (
        os.path.exists(os.path.join(tmp_path, "cls_forest_smpls_bal_train.h5"))
        and os.path.exists(os.path.join(tmp_path, "cls_grass_smpls_bal_test.h5"))
        and os.path.exists(os.path.join(tmp_path, "cls_urban_smpls_bal_valid.h5"))
        and os.path.exists(os.path.join(tmp_path, "cls_water_smpls_bal_train.h5"))
    )


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_split_chip_sample_train_valid_test(tmp_path):
    import rsgislib.classification

    in_h5_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_chip_vals.h5"
    )

    train_h5_file = os.path.join(tmp_path, "out_train_data.h5")
    valid_h5_file = os.path.join(tmp_path, "out_valid_data.h5")
    test_h5_file = os.path.join(tmp_path, "out_test_data.h5")

    rsgislib.classification.split_chip_sample_train_valid_test(
        in_h5_file,
        train_h5_file,
        valid_h5_file,
        test_h5_file,
        2,
        2,
        train_sample=None,
        rnd_seed=42,
        datatype=rsgislib.TYPE_16INT,
    )

    assert (
        os.path.exists(train_h5_file)
        and os.path.exists(valid_h5_file)
        and os.path.exists(test_h5_file)
    )


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_split_chip_sample_ref_train_valid_test(tmp_path):
    import rsgislib.classification

    in_h5_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_refchip_vals.h5"
    )

    train_h5_file = os.path.join(tmp_path, "out_train_data.h5")
    valid_h5_file = os.path.join(tmp_path, "out_valid_data.h5")
    test_h5_file = os.path.join(tmp_path, "out_test_data.h5")

    rsgislib.classification.split_chip_sample_ref_train_valid_test(
        in_h5_file,
        train_h5_file,
        valid_h5_file,
        test_h5_file,
        2,
        2,
        train_sample=None,
        rnd_seed=42,
        datatype=rsgislib.TYPE_16INT,
    )

    assert (
        os.path.exists(train_h5_file)
        and os.path.exists(valid_h5_file)
        and os.path.exists(test_h5_file)
    )


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_flip_chip_hdf5_file(tmp_path):
    import rsgislib.classification

    in_h5_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_chip_vals.h5"
    )

    out_h5_file = os.path.join(tmp_path, "out_data.h5")

    rsgislib.classification.flip_chip_hdf5_file(
        in_h5_file, out_h5_file, datatype=rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_file)


@pytest.mark.skipif(H5PY_NOT_AVAIL, reason="h5py dependency not available")
def test_flip_ref_chip_hdf5_file(tmp_path):
    import rsgislib.classification

    in_h5_file = os.path.join(
        ZONALSTATS_DATA_DIR, "sen2_20210527_aber_b1-6_refchip_vals.h5"
    )

    out_h5_file = os.path.join(tmp_path, "out_data.h5")

    rsgislib.classification.flip_ref_chip_hdf5_file(
        in_h5_file, out_h5_file, datatype=rsgislib.TYPE_16INT
    )

    assert os.path.exists(out_h5_file)


def test_get_class_training_data(tmp_path):
    import rsgislib.imageutils
    import rsgislib.classification

    s2_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    cls_vec_smpls = os.path.join(CLASSIFICATION_DATA_DIR, "cls_poly_smpls.gpkg")

    img_band_info = []
    img_band_info.append(
        rsgislib.imageutils.ImageBandInfo(s2_img, "s2", [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    )

    class_vec_sample_info = []
    class_vec_sample_info.append(
        rsgislib.classification.ClassVecSamplesInfoObj(
            id=1,
            class_name="Forest",
            vec_file=cls_vec_smpls,
            vec_lyr="cls_forest_smpls",
            file_h5=os.path.join(tmp_path, "cls_forest_smpls.h5"),
        )
    )
    class_vec_sample_info.append(
        rsgislib.classification.ClassVecSamplesInfoObj(
            id=2,
            class_name="Grass",
            vec_file=cls_vec_smpls,
            vec_lyr="cls_grass_smpls",
            file_h5=os.path.join(tmp_path, "cls_grass_smpls.h5"),
        )
    )
    class_vec_sample_info.append(
        rsgislib.classification.ClassVecSamplesInfoObj(
            id=3,
            class_name="Urban",
            vec_file=cls_vec_smpls,
            vec_lyr="cls_urban_smpls",
            file_h5=os.path.join(tmp_path, "cls_urban_smpls.h5"),
        )
    )
    class_vec_sample_info.append(
        rsgislib.classification.ClassVecSamplesInfoObj(
            id=4,
            class_name="Water",
            vec_file=cls_vec_smpls,
            vec_lyr="cls_water_smpls",
            file_h5=os.path.join(tmp_path, "cls_water_smpls.h5"),
        )
    )

    cls_info = rsgislib.classification.get_class_training_data(
        img_band_info,
        class_vec_sample_info,
        tmp_dir=tmp_path,
        sub_sample=None,
        ref_img=s2_img,
    )

    all_files_present = True
    for cls in cls_info:
        if not os.path.exists(cls_info[cls].file_h5):
            all_files_present = False

    assert all_files_present


"""
@pytest.mark.skipif(
    (PLOTLY_NOT_AVAIL or KALEIDO_NOT_AVAIL),
    reason="plotly or kaleido dependencies not available",
)
def test_plot_train_data(tmp_path):
    import glob
    import rsgislib.classification

    forest_h5 = os.path.join(CLASSIFICATION_DATA_DIR, "cls_forest_smpls.h5")
    urban_h5 = os.path.join(CLASSIFICATION_DATA_DIR, "cls_urban_smpls.h5")

    rsgislib.classification.plot_train_data(
        forest_h5,
        urban_h5,
        tmp_path,
        cls1_name="Forest",
        cls2_name="Urban",
        var_names=None,
    )

    plot_files = glob.glob(os.path.join(tmp_path, "*.png"))

    assert len(plot_files) == 100
"""


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_create_acc_pt_sets(tmp_path):
    import rsgislib.classification
    import glob

    vec_file = os.path.join(
        CLASSIFICATION_DATA_DIR, "gmw_acc_roi_30_cls_acc_pts.geojson"
    )
    vec_lyr = "gmw_acc_roi_30_cls_acc_pts"

    out_vec_file_base = os.path.join(tmp_path, "gmw_acc_roi_30_cls_acc_pts_")

    rsgislib.classification.create_acc_pt_sets(
        vec_file,
        vec_lyr,
        out_vec_file_base,
        "gmw_acc_roi_30_cls_acc_pts",
        "gmw_v25_cls",
        10,
        sets_col="set_id",
        out_format="GeoJSON",
        out_ext="geojson",
        shuffle_rows=True,
        rnd_seed=42,
    )
    assert (
        len(glob.glob(os.path.join(tmp_path, "gmw_acc_roi_30_cls_acc_pts_*.geojson")))
        == 10
    )


def test_generate_random_accuracy_pts(tmp_path):
    import rsgislib.classification

    input_img = os.path.join(CLASSIFICATION_DATA_DIR, "gmw_acc_roi_1_cls.kea")

    out_vec_file = os.path.join(tmp_path, "out_vecs.gpkg")
    out_vec_lyr = "out_vecs"

    rsgislib.classification.generate_random_accuracy_pts(
        input_img,
        out_vec_file,
        out_vec_lyr,
        "GPKG",
        "cls_name",
        "gmw_v2_cls",
        "ref_cls",
        1000,
        42,
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_generate_stratified_random_accuracy_pts(tmp_path):
    import rsgislib.classification

    input_img = os.path.join(CLASSIFICATION_DATA_DIR, "gmw_acc_roi_1_cls.kea")

    out_vec_file = os.path.join(tmp_path, "out_vecs.gpkg")
    out_vec_lyr = "out_vecs"

    rsgislib.classification.generate_stratified_random_accuracy_pts(
        input_img,
        out_vec_file,
        out_vec_lyr,
        "GPKG",
        "cls_name",
        "gmw_v2_cls",
        "ref_cls",
        1000,
        42,
        False,
        True,
    )

    assert os.path.exists(out_vec_file)


def test_pop_class_info_accuracy_pts_only_cls_col(tmp_path):
    import rsgislib.classification

    input_img = os.path.join(CLASSIFICATION_DATA_DIR, "gmw_acc_roi_1_cls.kea")

    vec_ref_file = os.path.join(
        CLASSIFICATION_DATA_DIR, "gmw_acc_set_1_acc_pts.geojson"
    )
    vec_file = os.path.join(tmp_path, "gmw_acc_set_1_acc_pts.geojson")
    copy2(vec_ref_file, vec_file)
    vec_lyr = "gmw_acc_set_1_acc_pts"

    rsgislib.classification.pop_class_info_accuracy_pts(
        input_img, vec_file, vec_lyr, "cls_name", "gmw_new_cls"
    )


def test_pop_class_info_accuracy_pts_all_col(tmp_path):
    import rsgislib.classification

    input_img = os.path.join(CLASSIFICATION_DATA_DIR, "gmw_acc_roi_1_cls.kea")

    vec_ref_file = os.path.join(
        CLASSIFICATION_DATA_DIR, "gmw_acc_set_1_acc_pts.geojson"
    )
    vec_file = os.path.join(tmp_path, "gmw_acc_set_1_acc_pts.geojson")
    copy2(vec_ref_file, vec_file)
    vec_lyr = "gmw_acc_set_1_acc_pts"

    rsgislib.classification.pop_class_info_accuracy_pts(
        input_img,
        vec_file,
        vec_lyr,
        "cls_name",
        "gmw_new_cls",
        "ref_new_col",
        "new_process",
    )


def test_get_class_info_dict(tmp_path):
    import rsgislib.classification

    cls_in_info = dict()
    cls_in_info["Forest"] = rsgislib.classification.ClassSimpleInfoObj(
        id=1,
        file_h5=os.path.join(CLASSIFICATION_DATA_DIR, "cls_forest_smpls_bal.h5"),
        red=120,
        green=120,
        blue=120,
    )
    cls_in_info["Grass"] = rsgislib.classification.ClassSimpleInfoObj(
        id=2,
        file_h5=os.path.join(CLASSIFICATION_DATA_DIR, "cls_grass_smpls_bal.h5"),
        red=120,
        green=120,
        blue=120,
    )
    cls_in_info["Urban"] = rsgislib.classification.ClassSimpleInfoObj(
        id=3,
        file_h5=os.path.join(CLASSIFICATION_DATA_DIR, "cls_urban_smpls_bal.h5"),
        red=120,
        green=120,
        blue=120,
    )
    cls_in_info["Water"] = rsgislib.classification.ClassSimpleInfoObj(
        id=4,
        file_h5=os.path.join(CLASSIFICATION_DATA_DIR, "cls_water_smpls_bal.h5"),
        red=120,
        green=120,
        blue=120,
    )

    cls_out_info = rsgislib.classification.get_class_info_dict(
        cls_in_info, smpls_dir=tmp_path
    )



#!/usr/bin/env python
"""
The tools.sysprofile module contains functions which can be used to profile the memory and CPU usage of a system.
These are basic tools and do not replace more robust profiling tools.
"""


def create_mem_cpu_profile(out_file, interval=3, duration=600):
    """
    A function which creates a profile of memory and cpu usage of the system.

    This function provides information for the whole system not a single process. Therefore, if you are trying
    to study the performance of a single function you should minimise the use of the system for anything other
    than the experiment you are performing in-order to avoid misleading results.

    It is recommended that you run the function for a period (e.g., 30 seconds) before starting your processing
    so the background system usage can be estimated when the results are analysed.

    :param out_file: The output file to which the information will be saved.
    :param interval: The interval, in seconds, at which the information
                     should be recorded. (default: 3 seconds)
    :param duration: The time, in seconds, for which the function should
                     run for. (default: 600 seconds; 10 mins)

    """
    import psutil
    import time
    import tqdm

    n_steps = int(duration / interval)
    psutil.cpu_percent()
    time.sleep(1)
    print("Started logging...")
    with open(out_file, "w") as out_file_obj:
        start_time = time.time()
        out_file_obj.write("timestamp,interval,used_mem,cpu_percent\n")
        for i in tqdm.tqdm(range(n_steps)):
            current_time = time.time()
            current_time_inter = str(
                round(current_time - start_time, 1)
            )  # time in seconds
            mem_info = dict(psutil.virtual_memory()._asdict())
            used_mem = str(round(mem_info["used"] * 0.000000001, 2))  # mem usage in GB
            out_file_obj.write(
                "{}, {}, {}, {}\n".format(
                    current_time, current_time_inter, used_mem, psutil.cpu_percent()
                )
            )
            out_file_obj.flush()
            time.sleep(interval)


def plot_mem_cpu_profile(
    profile_file,
    out_plot_file="sys_profile.pdf",
    ref_period=30,
    interval=3,
    plot_format="PDF",
):
    """
    A function which creates a plot using the output of the rsgislib.tools.create_mem_cpu_profile function.
    A reference period can be used to estimate the background system usage which is removed from the whole
    timeseries.

    :param profile_file: a csv file from the rsgislib.tools.create_mem_cpu_profile function.
    :param out_plot_file: the output image file. Default: sys_profile.pdf
    :param ref_period: a reference period (in seconds) from which the background system resources will
                       be estimated and removed from the whole timeseries. Default: 30 seconds. If a value of
                       0 is provided (or less 4 x interval) then no reference period will be used.
    :param interval: the interval (in seconds) between measurements; used within the create_mem_cpu_profile function.
    :param plot_format: the file format for the outputted plot (i.e., PDF or PNG). Default: PDF

    """

    import pandas
    import matplotlib.pyplot as plt

    cols_dtypes = {
        "timestamp": float,
        "interval": float,
        "used_mem": float,
        "cpu_percent": float,
    }
    profile_df = pandas.read_csv(
        profile_file, delimiter=",", header=0, dtype=cols_dtypes
    )

    if (ref_period > 0) and (ref_period > (4 * interval)):
        ref_n_rows = int(ref_period / interval)
        avg_bkg_used_mem = profile_df[0:ref_n_rows].mean()["used_mem"]
        avg_bkg_cpu_percent = profile_df[0:ref_n_rows].mean()["cpu_percent"]

        profile_df["used_mem"] = profile_df["used_mem"] - avg_bkg_used_mem
        profile_df["cpu_percent"] = profile_df["cpu_percent"] - avg_bkg_cpu_percent
        profile_df["used_mem"][profile_df["used_mem"] < 0] = 0.0
        profile_df["cpu_percent"][profile_df["cpu_percent"] < 0] = 0.0

    time_max = profile_df["interval"].max()
    used_mem_max = profile_df["used_mem"].max()
    cpu_percent_max = profile_df["cpu_percent"].max()

    ax = profile_df.plot(x="interval", y="used_mem", color="r", label="Used Memory")
    profile_df.plot(
        x="interval",
        y="cpu_percent",
        color="b",
        label="Percent CPU",
        secondary_y=True,
        ax=ax,
    )

    ax.set_xlabel("time (seconds)")
    ax.set_xlim(0, time_max)
    ax.set_ylabel("Memory (Gb)")
    ax.set_ylim(0, used_mem_max * 1.10)
    ax.right_ax.set_ylabel("CPU Percent (%)")
    ax.right_ax.set_ylim(0, cpu_percent_max * 1.10)
    plt.savefig(out_plot_file, format=plot_format)



import os

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def test_stdimgblockiter_ndvi(tmp_path):
    import rsgislib.imageutils
    import rsgislib.imageutils.stdimgblockiter
    import numpy
    import tqdm

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_ndvi.kea")

    imgBandInfo = []
    imgBandInfo.append(
        rsgislib.imageutils.ImageBandInfo(
            file_name=input_img, name="sen2", bands=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        )
    )

    imgOutInfo = []
    imgOutInfo.append(
        rsgislib.imageutils.OutImageInfo(
            file_name=output_img,
            name="sen2_ndvi",
            nbands=1,
            no_data_val=-999,
            gdalformat="KEA",
            datatype=rsgislib.TYPE_32FLOAT,
        )
    )

    img_iter = rsgislib.imageutils.stdimgblockiter.StdImgBlockIter(imgBandInfo, 64)
    img_iter.create_output_imgs(imgOutInfo)
    for x_block, y_block, out_img_data in tqdm.tqdm(img_iter):
        ndvi = (
            out_img_data["sen2"]["np_arr"][8] - out_img_data["sen2"]["np_arr"][3]
        ) / (out_img_data["sen2"]["np_arr"][8] + out_img_data["sen2"]["np_arr"][3])
        out_data = {"sen2_ndvi": numpy.expand_dims(ndvi, axis=0)}
        img_iter.write_block_to_image(x_block, y_block, out_data)

    assert os.path.exists(output_img)


def test_stdimgblockiter_ndvi_ndwi(tmp_path):
    import rsgislib.imageutils
    import rsgislib.imageutils.stdimgblockiter
    import numpy
    import tqdm

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_ndvi_img = os.path.join(tmp_path, "sen2_20210527_aber_ndvi.kea")
    out_ndwi_img = os.path.join(tmp_path, "sen2_20210527_aber_ndwi.kea")

    imgBandInfo = []
    imgBandInfo.append(
        rsgislib.imageutils.ImageBandInfo(
            file_name=input_img, name="sen2", bands=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        )
    )

    imgOutInfo = []
    imgOutInfo.append(
        rsgislib.imageutils.OutImageInfo(
            file_name=out_ndvi_img,
            name="sen2_ndvi",
            nbands=1,
            no_data_val=-999,
            gdalformat="KEA",
            datatype=rsgislib.TYPE_32FLOAT,
        )
    )
    imgOutInfo.append(
        rsgislib.imageutils.OutImageInfo(
            file_name=out_ndwi_img,
            name="sen2_ndwi",
            nbands=1,
            no_data_val=-999,
            gdalformat="KEA",
            datatype=rsgislib.TYPE_32FLOAT,
        )
    )

    img_iter = rsgislib.imageutils.stdimgblockiter.StdImgBlockIter(imgBandInfo, 64)
    img_iter.create_output_imgs(imgOutInfo)
    for x_block, y_block, out_img_data in tqdm.tqdm(img_iter):
        ndvi = (
            out_img_data["sen2"]["np_arr"][8] - out_img_data["sen2"]["np_arr"][3]
        ) / (out_img_data["sen2"]["np_arr"][8] + out_img_data["sen2"]["np_arr"][3])
        ndwi = (
            out_img_data["sen2"]["np_arr"][8] - out_img_data["sen2"]["np_arr"][9]
        ) / (out_img_data["sen2"]["np_arr"][8] + out_img_data["sen2"]["np_arr"][9])
        out_data = {
            "sen2_ndvi": numpy.expand_dims(ndvi, axis=0),
            "sen2_ndwi": numpy.expand_dims(ndwi, axis=0),
        }
        img_iter.write_block_to_image(x_block, y_block, out_data)

    assert os.path.exists(out_ndvi_img) and os.path.exists(out_ndwi_img)



#! /usr/bin/env python
############################################################################
#  clustersklearn.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 09/01/2020
# Version: 1.0
#
# History:
# Version 1.0 - Refactored from previous structure.
#
###########################################################################

import rsgislib
import rsgislib.imageutils
import rsgislib.rastergis

import tqdm

import numpy

from osgeo import gdal

from sklearn.cluster import MiniBatchKMeans

from rios.imagereader import ImageReader
from rios.imagewriter import ImageWriter


def img_pixel_sample_cluster(
    input_img,
    output_img,
    gdalformat="KEA",
    no_data_val=0,
    n_img_smpl=100,
    clusterer=MiniBatchKMeans(
        n_clusters=60, init="k-means++", max_iter=100, batch_size=100
    ),
    calc_stats=True,
    use_mean_shift_est_band_width=False,
):
    """
    A function which allows a clustering to be performed using the algorithms available
    within the scikit-learn library. The clusterer is trained on a sample of the input
    image and then applied using the predict function (therefore this function is only
    compatible with clusterers which have the predict function implemented) to the
    whole image.

    :param input_img: input image file.
    :param output_img: output image file.
    :param gdalformat: output image file format.
    :param no_data_val: no data value associated with the input image.
    :param n_img_smpl: the input image sampling. (e.g., 100 is every 100th pixel)
    :param clusterer: clusterer from scikit-learn which must have a predict function.
    :param calc_stats: calculate image pixel statistics, histogram and image pyramids -
                       note if you are not using a KEA file then the format needs to
                       support RATs for this option as histogram and colour table
                       are written to RAT.
    :param use_mean_shift_est_band_width: use the mean-shift algorithm as the clusterer
                                          (pass None as the clusterer) where the
                                          bandwidth is calculated from the data itself.
    """
    print("Sample input image:")
    dataSamp = rsgislib.imageutils.extract_img_pxl_sample(
        input_img, n_img_smpl, no_data_val
    )

    if use_mean_shift_est_band_width:
        print("Using Mean-Shift predict bandwidth")
        from sklearn.cluster import MeanShift, estimate_bandwidth

        bandwidth = estimate_bandwidth(dataSamp, quantile=0.2, n_samples=500)
        clusterer = MeanShift(bandwidth=bandwidth, bin_seeding=True)

    print("Fit Clusterer")
    clusterer.fit(dataSamp)
    print("Fitted Clusterer")

    print("Apply to whole image:")
    reader = ImageReader(input_img, windowxsize=200, windowysize=200)
    writer = None
    for (info, block) in tqdm.tqdm(reader):
        blkShape = block.shape
        blkBands = block.reshape((blkShape[0], (blkShape[1] * blkShape[2]))).T
        ID = numpy.arange(blkBands.shape[0])
        outClusterVals = numpy.zeros((blkBands.shape[0]))

        finiteMskArr = numpy.isfinite(blkBands).all(axis=1)
        ID = ID[finiteMskArr]
        blkBands = blkBands[finiteMskArr]

        noDataValArr = numpy.logical_not(
            numpy.where(blkBands == no_data_val, True, False).all(axis=1)
        )

        blkBandsNoData = blkBands[noDataValArr]
        ID = ID[noDataValArr]

        if ID.shape[0] > 0:
            outPred = clusterer.predict(blkBandsNoData) + 1
            outClusterVals[ID] = outPred

        outClusterValsOutArr = outClusterVals.reshape([1, blkShape[1], blkShape[2]])

        if writer is None:
            writer = ImageWriter(
                output_img,
                info=info,
                firstblock=outClusterValsOutArr,
                drivername=gdalformat,
                creationoptions=[],
            )
        else:
            writer.write(outClusterValsOutArr)
    writer.close(calcStats=False)

    if calc_stats:
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )


def img_pixel_tiled_cluster(
    input_img,
    output_img,
    gdalformat="KEA",
    no_data_val=0,
    clusterer=MiniBatchKMeans(
        n_clusters=60, init="k-means++", max_iter=100, batch_size=100
    ),
    calc_stats=True,
    use_mean_shift_est_band_width=False,
    tile_x_size=200,
    tile_y_size=200,
):
    """
    A function which allows a clustering to be performed using the algorithms available
    within the scikit-learn library. The clusterer is applied to a single tile at a time
    and therefore produces tile boundaries in the result. However, memory is controlled
    such that usage isn't excessive which it could be when processing a whole image.

    :param input_img: input image file.
    :param output_img: output image file.
    :param gdalformat: output image file format.
    :param no_data_val: no data value associated with the input image.
    :param clusterer: clusterer from scikit-learn which must have a predict function.
    :param calc_stats: calculate image pixel statistics, histogram and image pyramids -
                       note if you are not using a KEA file then the format needs to
                       support RATs for this option as histogram and colour table
                       are written to RAT.
    :param use_mean_shift_est_band_width: use the mean-shift algorithm as the clusterer
                                          (pass None as the clusterer) where the
                                          bandwidth is calculated from the data itself.
    :param tile_x_size: tile size in the x-axis in pixels.
    :param tile_y_size: tile size in the y-axis in pixels.
    """
    if use_mean_shift_est_band_width:
        from sklearn.cluster import MeanShift, estimate_bandwidth

    reader = ImageReader(input_img, windowxsize=tile_x_size, windowysize=tile_y_size)
    writer = None
    for (info, block) in tqdm.tqdm(reader):
        blkShape = block.shape
        blkBands = block.reshape((blkShape[0], (blkShape[1] * blkShape[2]))).T
        ID = numpy.arange(blkBands.shape[0])
        outClusterVals = numpy.zeros((blkBands.shape[0]))

        finiteMskArr = numpy.isfinite(blkBands).all(axis=1)
        ID = ID[finiteMskArr]
        blkBands = blkBands[finiteMskArr]

        noDataValArr = numpy.logical_not(
            numpy.where(blkBands == no_data_val, True, False).all(axis=1)
        )

        blkBandsNoData = blkBands[noDataValArr]
        ID = ID[noDataValArr]

        if ID.shape[0] > 0:
            if use_mean_shift_est_band_width:
                bandwidth = estimate_bandwidth(
                    blkBandsNoData, quantile=0.2, n_samples=1000
                )
                clusterer = MeanShift(bandwidth=bandwidth, bin_seeding=True)

            clusterer.fit(blkBandsNoData)
            outPred = clusterer.labels_ + 1
            outClusterVals[ID] = outPred

        outClusterValsOutArr = outClusterVals.reshape([1, blkShape[1], blkShape[2]])

        if writer is None:
            writer = ImageWriter(
                output_img,
                info=info,
                firstblock=outClusterValsOutArr,
                drivername=gdalformat,
                creationoptions=[],
            )
        else:
            writer.write(outClusterValsOutArr)
    writer.close(calcStats=False)

    if calc_stats:
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )


def img_pixel_cluster(
    input_img,
    output_img,
    gdalformat="KEA",
    no_data_val=0,
    clusterer=MiniBatchKMeans(
        n_clusters=60, init="k-means++", max_iter=100, batch_size=100
    ),
    calc_stats=True,
    use_mean_shift_est_band_width=False,
):
    """
    A function which allows a clustering to be performed using the algorithms available
    within the scikit-learn library. The clusterer is applied to the whole image in one
    operation so therefore requires the whole image to be loaded into memory. However,
    if there is sufficient memory all the clustering algorithms within scikit-learn can
    be applied without boundary artifacts.

    :param input_img: input image file.
    :param output_img: output image file.
    :param gdalformat: output image file format.
    :param no_data_val: no data value associated with the input image.
    :param clusterer: clusterer from scikit-learn which must have a predict function.
    :param calc_stats: calculate image pixel statistics, histogram and image pyramids -
                       note if you are not using a KEA file then the format needs to
                       support RATs for this option as histogram and colour table
                       are written to RAT.
    :param use_mean_shift_est_band_width: use the mean-shift algorithm as the clusterer
                                          (pass None as the clusterer) where the
                                          bandwidth is calculated from the data itself.
    """
    # Create output image
    rsgislib.imageutils.create_copy_img(
        input_img, output_img, 1, 0, gdalformat, rsgislib.TYPE_16UINT
    )

    if use_mean_shift_est_band_width:
        from sklearn.cluster import MeanShift, estimate_bandwidth

    gdalDS = gdal.Open(input_img, gdal.GA_ReadOnly)
    nPxls = gdalDS.RasterXSize * gdalDS.RasterYSize

    pxlVals = numpy.zeros((gdalDS.RasterCount, nPxls))

    for nBand in numpy.arange(gdalDS.RasterCount):
        gdalBand = gdalDS.GetRasterBand(int(nBand + 1))
        imgArr = gdalBand.ReadAsArray().flatten()
        pxlVals[nBand] = imgArr

    pxlVals = pxlVals.T

    ID = numpy.arange(pxlVals.shape[0])
    outClusterVals = numpy.zeros((pxlVals.shape[0]))

    finiteMskArr = numpy.isfinite(pxlVals).all(axis=1)
    ID = ID[finiteMskArr]
    pxlVals = pxlVals[finiteMskArr]

    noDataValArr = numpy.logical_not(
        numpy.where(pxlVals == no_data_val, True, False).all(axis=1)
    )

    pxlVals = pxlVals[noDataValArr]
    ID = ID[noDataValArr]

    if ID.shape[0] > 0:
        if use_mean_shift_est_band_width:
            bandwidth = estimate_bandwidth(pxlVals, quantile=0.2, n_samples=1000)
            clusterer = MeanShift(bandwidth=bandwidth, bin_seeding=True)
        print("Perform Clustering")
        clusterer.fit(pxlVals)
        print("Performed Clustering")
        outPred = clusterer.labels_ + 1
        outClusterVals[ID] = outPred

    outClusterValsOutArr = outClusterVals.reshape(
        [gdalDS.RasterYSize, gdalDS.RasterXSize]
    )
    print(outClusterValsOutArr.shape)

    gdalOutDS = gdal.Open(output_img, gdal.GA_Update)
    gdalOutBand = gdalOutDS.GetRasterBand(1)
    gdalOutBand.WriteArray(outClusterValsOutArr)
    gdalOutDS = None
    gdalDS = None

    if calc_stats:
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )
